import asyncio
import logging
import secrets
import uuid
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from datetime import datetime
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Optional, Self

import aiohttp
import aiohttp_cors
import aiotools
import attrs
import sqlalchemy as sa
from aiohttp import web
from pydantic import (
    AliasChoices,
    AnyUrl,
    ConfigDict,
    Field,
    HttpUrl,
    NonNegativeFloat,
    NonNegativeInt,
    model_validator,
)
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload
from sqlalchemy.orm.exc import NoResultFound
from yarl import URL

from ai.backend.common import typed_validators as tv
from ai.backend.common.api_handlers import BaseFieldModel
from ai.backend.common.bgtask import ProgressReporter
from ai.backend.common.events import (
    EventHandler,
    KernelLifecycleEventReason,
    ModelServiceStatusEvent,
    SessionCancelledEvent,
    SessionEnqueuedEvent,
    SessionPreparingEvent,
    SessionStartedEvent,
    SessionTerminatedEvent,
)
from ai.backend.common.json import dump_json_str
from ai.backend.common.types import (
    MODEL_SERVICE_RUNTIME_PROFILES,
    AccessKey,
    AgentId,
    ClusterMode,
    ImageAlias,
    RuntimeVariant,
    SessionTypes,
    VFolderMount,
    VFolderUsageMode,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.models.image import ImageIdentifier
from ai.backend.manager.models.vfolder import VFolderOwnershipType, VFolderRow

from ..defs import DEFAULT_IMAGE_ARCH
from ..models import (
    EndpointLifecycle,
    EndpointRow,
    EndpointTokenRow,
    ImageRow,
    KernelLoadingStrategy,
    KeyPairRow,
    ModelServicePredicateChecker,
    RouteStatus,
    RoutingRow,
    SessionRow,
    UserRole,
    UserRow,
    query_accessible_vfolders,
    resolve_group_name_or_id,
    scaling_groups,
    vfolders,
)
from ..types import MountOptionModel, UserScope
from .auth import auth_required
from .exceptions import InvalidAPIParameters, ObjectNotFound, VFolderNotFound
from .manager import ALL_ALLOWED, READ_ALLOWED, server_status_required
from .session import query_userinfo
from .types import CORSOptions, WebMiddleware
from .utils import (
    LegacyBaseRequestModel,
    LegacyBaseResponseModel,
    get_access_key_scopes,
    get_user_uuid_scopes,
    pydantic_params_api_handler,
    pydantic_response_api_handler,
    undefined,
)

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__name__))


async def is_user_allowed_to_access_resource(
    db_sess: AsyncSession,
    request: web.Request,
    resource_owner: uuid.UUID,
) -> bool:
    if request["user"]["is_superadmin"]:
        return True
    elif request["user"]["is_admin"]:
        query = sa.select(UserRow).filter(UserRow.uuid == resource_owner)
        result = await db_sess.execute(query)
        user = result.scalar()
        return user.domain_name == request["user"]["domain_name"]
    else:
        return request["user"]["uyud"] == resource_owner


class ListServeRequestModel(LegacyBaseRequestModel):
    name: str | None = Field(default=None)


class SuccessResponseModel(LegacyBaseResponseModel):
    success: bool = Field(default=True)


class CompactServeInfoModel(LegacyBaseResponseModel):
    id: uuid.UUID = Field(description="Unique ID referencing the model service.")
    name: str = Field(description="Name of the model service.")
    replicas: NonNegativeInt = Field(description="Number of identical inference sessions.")
    desired_session_count: NonNegativeInt = Field(description="Deprecated; use `replicas` instead.")
    active_route_count: NonNegativeInt = Field(
        description=(
            "Information of routes which are actually spawned and ready to accept the traffic."
        )
    )
    service_endpoint: HttpUrl | None = Field(
        default=None,
        description=(
            "HTTP(S) endpoint to the API service. This field will be filed after the attempt to"
            " create a first inference session succeeds. Endpoint created is fixed and immutable"
            " for the bound endpoint until the endpoint is destroyed."
        ),
    )
    is_public: bool = Field(
        description=(
            'Indicates if the API endpoint is open to public. In this context "public" means there'
            " will be no authentication required to communicate with this API service."
        )
    )


@auth_required
@server_status_required(READ_ALLOWED)
@pydantic_params_api_handler(ListServeRequestModel)
async def list_serve(
    request: web.Request, params: ListServeRequestModel
) -> list[CompactServeInfoModel]:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]

    log.info("SERVE.LIST (email:{}, ak:{})", request["user"]["email"], access_key)
    query_conds = (EndpointRow.session_owner == request["user"]["uuid"]) & (
        EndpointRow.lifecycle_stage == EndpointLifecycle.CREATED
    )
    if params.name:
        query_conds &= EndpointRow.name == params.name

    async with root_ctx.db.begin_readonly_session() as db_sess:
        query = (
            sa.select(EndpointRow).where(query_conds).options(selectinload(EndpointRow.routings))
        )
        result = await db_sess.execute(query)
        rows = result.scalars().all()

    return [
        CompactServeInfoModel(
            id=endpoint.id,
            name=endpoint.name,
            replicas=endpoint.replicas,
            desired_session_count=endpoint.replicas,
            active_route_count=len([
                r for r in endpoint.routings if r.status == RouteStatus.HEALTHY
            ]),
            service_endpoint=endpoint.url,
            is_public=endpoint.open_to_public,
        )
        for endpoint in rows
    ]


class RouteInfoModel(BaseFieldModel):
    route_id: uuid.UUID = Field(
        description=(
            "Unique ID referencing endpoint route. Each endpoint route has a one-to-one"
            " relationship with the inference session."
        )
    )
    session_id: uuid.UUID = Field(description="Unique ID referencing the inference session.")
    traffic_ratio: NonNegativeFloat


class ServeInfoModel(LegacyBaseResponseModel):
    model_config = ConfigDict(protected_namespaces=())

    endpoint_id: uuid.UUID = Field(description="Unique ID referencing the model service.")
    model_id: uuid.UUID = Field(description="ID of model VFolder.")
    extra_mounts: Sequence[uuid.UUID] = Field(
        description="List of extra VFolders which will be mounted to model service session."
    )
    name: str = Field(description="Name of the model service.")
    replicas: NonNegativeInt = Field(description="Number of identical inference sessions.")
    desired_session_count: NonNegativeInt = Field(description="Deprecated; use `replicas` instead.")
    model_definition_path: str | None = Field(
        description="Path to the the model definition file. If not set, Backend.AI will look for model-definition.yml or model-definition.yaml by default."
    )
    active_routes: list[RouteInfoModel] = Field(
        description="Information of routes which are bound with healthy sessions."
    )
    service_endpoint: HttpUrl | None = Field(
        default=None,
        description=(
            "HTTP(S) endpoint to the API service. This field will be filed after the attempt to"
            " create a first inference session succeeds. Endpoint created is fixed and immutable"
            " for the bound endpoint until the endpoint is destroyed."
        ),
    )
    is_public: bool = Field(
        description=(
            'Indicates if the API endpoint is open to public. In this context "public" means there'
            " will be no authentication required to communicate with this API service."
        )
    )
    runtime_variant: RuntimeVariant = Field(
        description="Type of the inference runtime the image will try to load."
    )


@auth_required
@server_status_required(READ_ALLOWED)
@pydantic_response_api_handler
async def get_info(request: web.Request) -> ServeInfoModel:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info(
        "SERVE.GET_INFO (email:{}, ak:{}, s:{})", request["user"]["email"], access_key, service_id
    )

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound

    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})

    return ServeInfoModel(
        endpoint_id=endpoint.id,
        model_id=endpoint.model,
        extra_mounts=[m.vfid.folder_id for m in endpoint.extra_mounts],
        name=endpoint.name,
        model_definition_path=endpoint.model_definition_path,
        replicas=endpoint.replicas,
        desired_session_count=endpoint.replicas,
        active_routes=[
            RouteInfoModel(route_id=r.id, session_id=r.session, traffic_ratio=r.traffic_ratio)
            for r in endpoint.routings
            if r.status == RouteStatus.HEALTHY
        ],
        service_endpoint=endpoint.url,
        is_public=endpoint.open_to_public,
        runtime_variant=endpoint.runtime_variant,
    )


class ServiceConfigModel(LegacyBaseRequestModel):
    model_config = ConfigDict(protected_namespaces=())

    model: str = Field(description="Name or ID of the model VFolder", examples=["ResNet50"])
    model_definition_path: str | None = Field(
        description="Path to the model definition file. If not set, Backend.AI will look for model-definition.yml or model-definition.yaml by default.",
        default=None,
    )
    model_version: int = Field(
        description="Unused; Reserved for future works",
        default=1,
        alias="modelVersion",
    )
    model_mount_destination: str = Field(
        default="/models",
        description=(
            "Mount destination for the model VFolder will be mounted inside the inference session. Must be set to `/models` when choosing `runtime_variant` other than `CUSTOM` or `CMD`."
        ),
        alias="modelMountDestination",
    )

    extra_mounts: dict[uuid.UUID, MountOptionModel] = Field(
        description=(
            "Specifications about extra VFolders mounted to model service session. "
            "MODEL type VFolders are not allowed to be attached to model service session with this option."
        ),
        default={},
    )

    environ: Optional[dict[str, str]] = Field(
        description="Environment variables to be set inside the inference session",
        default=None,
    )
    scaling_group: str = Field(
        description="Name of the resource group to spawn inference sessions",
        examples=["nvidia-H100"],
        alias="scalingGroup",
    )
    resources: dict[str, str | int] = Field(examples=[{"cpu": 4, "mem": "32g", "cuda.shares": 2.5}])
    resource_opts: dict[str, str | int] = Field(examples=[{"shmem": "2g"}], default={})


class NewServiceRequestModel(LegacyBaseRequestModel):
    service_name: tv.SessionName = Field(
        description="Name of the service",
        validation_alias=AliasChoices("name", "clientSessionToken"),
    )
    replicas: int = Field(
        description="Number of sessions to serve traffic. Replacement of `desired_session_count` (or `desiredSessionCount`).",
        validation_alias=AliasChoices("desired_session_count", "desiredSessionCount"),
    )
    image: str = Field(
        description="String reference of the image which will be used to create session",
        examples=["cr.backend.ai/stable/python-tensorflow:2.7-py38-cuda11.3"],
        alias="lang",
    )
    runtime_variant: RuntimeVariant = Field(
        description="Type of the inference runtime the image will try to load.",
        default=RuntimeVariant.CUSTOM,
    )
    architecture: str = Field(
        description="Image architecture",
        default=DEFAULT_IMAGE_ARCH,
        alias="arch",
    )
    group_name: str = Field(
        description="Name of project to spawn session",
        default="default",
        validation_alias=AliasChoices("group", "groupName"),
        serialization_alias="group",
    )
    domain_name: str = Field(
        description="Name of domain to spawn session",
        default="default",
        validation_alias=AliasChoices("domain", "domainName"),
        serialization_alias="domain",
    )
    cluster_size: int = Field(
        default=1,
        alias="clusterSize",
    )
    cluster_mode: str = Field(
        default="SINGLE_NODE",
        alias="clusterMode",
    )
    tag: str | None = Field(default=None)
    startup_command: str | None = Field(
        default=None,
        alias="startupCommand",
    )
    bootstrap_script: str | None = Field(
        default=None,
        alias="bootstrapScript",
    )
    callback_url: AnyUrl | None = Field(
        default=None,
        validation_alias=AliasChoices("callbackUrl", "CallbackURL"),
    )
    owner_access_key: str | None = Field(
        description=(
            "(for privileged users only) when specified, transfer ownership of the inference"
            " session to specified user"
        ),
        default=None,
    )
    open_to_public: bool = Field(
        description="If set to true, do not require an API key to access the model service",
        default=False,
    )
    config: ServiceConfigModel


@dataclass
class ValidationResult:
    model_id: uuid.UUID
    model_definition_path: str | None
    requester_access_key: AccessKey
    owner_access_key: AccessKey
    owner_uuid: uuid.UUID
    owner_role: UserRole
    group_id: uuid.UUID
    resource_policy: dict
    scaling_group: str
    extra_mounts: Sequence[VFolderMount]


async def _validate(request: web.Request, params: NewServiceRequestModel) -> ValidationResult:
    root_ctx: RootContext = request.app["_root.context"]
    scopes_param = {
        "owner_access_key": (
            None if params.owner_access_key is undefined else params.owner_access_key
        ),
    }

    requester_access_key, owner_access_key = await get_access_key_scopes(request, scopes_param)
    if params.replicas > (
        _m := request["user"]["resource_policy"]["max_session_count_per_model_session"]
    ):
        raise InvalidAPIParameters(f"Cannot spawn more than {_m} sessions for a single service")

    async with root_ctx.db.begin_readonly() as conn:
        checked_scaling_group = await ModelServicePredicateChecker.check_scaling_group(
            conn,
            params.config.scaling_group,
            owner_access_key,
            params.domain_name,
            params.group_name,
        )

        params.config.scaling_group = checked_scaling_group

        owner_uuid, group_id, resource_policy = await query_userinfo(
            request, params.model_dump(by_alias=True), conn
        )
        query = sa.select([UserRow.role]).where(UserRow.uuid == owner_uuid)
        owner_role = (await conn.execute(query)).scalar()
        assert owner_role

        allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
        try:
            extra_vf_conds = vfolders.c.id == uuid.UUID(params.config.model)
            matched_vfolders = await query_accessible_vfolders(
                conn,
                owner_uuid,
                user_role=owner_role,
                domain_name=params.domain_name,
                allowed_vfolder_types=allowed_vfolder_types,
                extra_vf_conds=extra_vf_conds,
            )
        except Exception as e:
            # just catching ValueError | VFolderNotFound will raise
            # TypeError: catching classes that do not inherit from BaseException is not allowed
            if isinstance(e, ValueError) or isinstance(e, VFolderNotFound):
                try:
                    extra_vf_conds = (vfolders.c.name == params.config.model) & (
                        vfolders.c.usage_mode == VFolderUsageMode.MODEL
                    )
                    matched_vfolders = await query_accessible_vfolders(
                        conn,
                        owner_uuid,
                        user_role=owner_role,
                        domain_name=params.domain_name,
                        allowed_vfolder_types=allowed_vfolder_types,
                        extra_vf_conds=extra_vf_conds,
                    )
                except VFolderNotFound as e:
                    raise VFolderNotFound("Cannot find model folder") from e
            else:
                raise
        if len(matched_vfolders) == 0:
            raise VFolderNotFound
        folder_row = matched_vfolders[0]
        if folder_row["usage_mode"] != VFolderUsageMode.MODEL:
            raise InvalidAPIParameters("Selected VFolder is not a model folder")

        model_id = folder_row["id"]

        vfolder_mounts = await ModelServicePredicateChecker.check_extra_mounts(
            conn,
            root_ctx.shared_config,
            root_ctx.storage_manager,
            model_id,
            params.config.model_mount_destination,
            params.config.extra_mounts,
            UserScope(
                domain_name=params.domain_name,
                group_id=group_id,
                user_uuid=owner_uuid,
                user_role=owner_role,
            ),
            resource_policy,
        )

    if params.runtime_variant == RuntimeVariant.CUSTOM:
        yaml_path = await ModelServicePredicateChecker.validate_model_definition(
            root_ctx.storage_manager,
            folder_row,
            params.config.model_definition_path,
        )
    else:
        if (
            params.runtime_variant != RuntimeVariant.CMD
            and params.config.model_mount_destination != "/models"
        ):
            raise InvalidAPIParameters(
                "Model mount destination must be /models for non-custom runtimes"
            )
        # this path won't be used on actual session but just to keep the convention
        yaml_path = "model-definition.yaml"

    return ValidationResult(
        model_id,
        yaml_path,
        requester_access_key,
        owner_access_key,
        owner_uuid,
        owner_role,
        group_id,
        resource_policy,
        checked_scaling_group,
        vfolder_mounts,
    )


@auth_required
@server_status_required(ALL_ALLOWED)
@pydantic_params_api_handler(NewServiceRequestModel)
async def create(request: web.Request, params: NewServiceRequestModel) -> ServeInfoModel:
    """
    Creates a new model service. If `replicas` is greater than zero,
    then inference sessions will be automatically scheduled upon successful creation of model service.
    """
    root_ctx: RootContext = request.app["_root.context"]

    validation_result = await _validate(request, params)

    async with root_ctx.db.begin_readonly_session() as session:
        model_vfolder_row = await VFolderRow.get(session, validation_result.model_id)
        if model_vfolder_row.ownership_type == VFolderOwnershipType.GROUP:
            raise InvalidAPIParameters(
                "Cannot create model service with the project type's vfolder"
            )

        image_row = await ImageRow.resolve(
            session,
            [
                ImageIdentifier(params.image, params.architecture),
                ImageAlias(params.image),
            ],
        )

    creation_config = params.config.model_dump()
    creation_config["mounts"] = [
        validation_result.model_id,
        *[m.vfid.folder_id for m in validation_result.extra_mounts],
    ]
    creation_config["mount_map"] = {
        validation_result.model_id: params.config.model_mount_destination,
        **{m.vfid.folder_id: m.kernel_path.as_posix() for m in validation_result.extra_mounts},
    }
    creation_config["mount_options"] = {
        m.vfid.folder_id: {"permission": m.mount_perm} for m in validation_result.extra_mounts
    }
    sudo_session_enabled = request["user"]["sudo_session_enabled"]

    # check if session is valid to be created
    await root_ctx.registry.create_session(
        "",
        image_row.image_ref,
        UserScope(
            domain_name=params.domain_name,
            group_id=validation_result.group_id,
            user_uuid=validation_result.owner_uuid,
            user_role=validation_result.owner_role,
        ),
        validation_result.owner_access_key,
        validation_result.resource_policy,
        SessionTypes.INFERENCE,
        creation_config,
        ClusterMode(params.cluster_mode),
        params.cluster_size,
        dry_run=True,  # Setting this to True will prevent actual session from being enqueued
        bootstrap_script=params.bootstrap_script,
        startup_command=params.startup_command,
        tag=params.tag,
        callback_url=URL(params.callback_url.unicode_string()) if params.callback_url else None,
        sudo_session_enabled=sudo_session_enabled,
    )

    async with root_ctx.db.begin_session() as db_sess:
        query = sa.select(EndpointRow).where(
            (EndpointRow.lifecycle_stage != EndpointLifecycle.DESTROYED)
            & (EndpointRow.name == params.service_name)
        )
        result = await db_sess.execute(query)
        service_with_duplicate_name = result.scalar()
        if service_with_duplicate_name is not None:
            raise InvalidAPIParameters("Cannot create multiple services with same name")

        project_id = await resolve_group_name_or_id(
            await db_sess.connection(), params.domain_name, params.group_name
        )
        if project_id is None:
            raise InvalidAPIParameters(f"Invalid group name {project_id}")
        endpoint = EndpointRow(
            params.service_name,
            validation_result.model_definition_path,
            request["user"]["uuid"],
            validation_result.owner_uuid,
            params.replicas,
            image_row,
            validation_result.model_id,
            params.domain_name,
            project_id,
            validation_result.scaling_group,
            params.config.resources,
            ClusterMode(params.cluster_mode),
            params.cluster_size,
            validation_result.extra_mounts,
            model_mount_destination=params.config.model_mount_destination,
            tag=params.tag,
            startup_command=params.startup_command,
            callback_url=URL(params.callback_url.unicode_string()) if params.callback_url else None,
            environ=params.config.environ,
            bootstrap_script=params.bootstrap_script,
            resource_opts=params.config.resource_opts,
            open_to_public=params.open_to_public,
            runtime_variant=params.runtime_variant,
        )
        db_sess.add(endpoint)
        await db_sess.flush()
        endpoint_id = endpoint.id

    return ServeInfoModel(
        endpoint_id=endpoint_id,
        model_id=endpoint.model,
        extra_mounts=[m.vfid.folder_id for m in endpoint.extra_mounts],
        name=params.service_name,
        model_definition_path=validation_result.model_definition_path,
        replicas=endpoint.replicas,
        desired_session_count=endpoint.replicas,
        active_routes=[],
        service_endpoint=None,
        is_public=params.open_to_public,
        runtime_variant=params.runtime_variant,
    )


class TryStartResponseModel(LegacyBaseResponseModel):
    task_id: str


@auth_required
@server_status_required(ALL_ALLOWED)
@pydantic_params_api_handler(NewServiceRequestModel)
async def try_start(request: web.Request, params: NewServiceRequestModel) -> TryStartResponseModel:
    root_ctx: RootContext = request.app["_root.context"]
    background_task_manager = root_ctx.background_task_manager

    validation_result = await _validate(request, params)

    async with root_ctx.db.begin_readonly_session() as session:
        image_row = await ImageRow.resolve(
            session,
            [
                ImageIdentifier(params.image, params.architecture),
                ImageAlias(params.image),
            ],
        )
        query = sa.select(sa.join(UserRow, KeyPairRow, KeyPairRow.user == UserRow.uuid)).where(
            UserRow.uuid == request["user"]["uuid"]
        )
        created_user = (await session.execute(query)).fetchone()

    creation_config = params.config.model_dump()
    creation_config["mount_map"] = {
        validation_result.model_id: params.config.model_mount_destination
    }
    sudo_session_enabled = request["user"]["sudo_session_enabled"]

    async def _task(reporter: ProgressReporter) -> None:
        terminated_event = asyncio.Event()

        result = await root_ctx.registry.create_session(
            f"model-eval-{secrets.token_urlsafe(16)}",
            image_row.image_ref,
            UserScope(
                domain_name=params.domain_name,
                group_id=validation_result.group_id,
                user_uuid=created_user.uuid,
                user_role=created_user.role,
            ),
            validation_result.owner_access_key,
            validation_result.resource_policy,
            SessionTypes.INFERENCE,
            {
                "mounts": [
                    validation_result.model_id,
                    *[m.vfid for m in validation_result.extra_mounts],
                ],
                "mount_map": {
                    validation_result.model_id: params.config.model_mount_destination,
                    **{m.vfid: m.kernel_path for m in validation_result.extra_mounts},
                },
                "mount_options": {
                    m.vfid: {"permission": m.mount_perm} for m in validation_result.extra_mounts
                },
                "model_definition_path": validation_result.model_definition_path,
                "environ": creation_config["environ"],
                "scaling_group": validation_result.scaling_group,
                "resources": creation_config["resources"],
                "resource_opts": creation_config["resource_opts"],
                "preopen_ports": None,
                "agent_list": None,
            },
            ClusterMode(params.cluster_mode),
            params.cluster_size,
            bootstrap_script=params.bootstrap_script,
            startup_command=params.startup_command,
            tag=params.tag,
            callback_url=URL(params.callback_url.unicode_string()) if params.callback_url else None,
            enqueue_only=True,
            sudo_session_enabled=sudo_session_enabled,
        )

        await reporter.update(
            message=dump_json_str({
                "event": "session_enqueued",
                "session_id": str(result["sessionId"]),
            })
        )

        async def _handle_event(
            context: None,
            source: AgentId,
            event: SessionEnqueuedEvent
            | SessionPreparingEvent
            | SessionStartedEvent
            | SessionCancelledEvent
            | SessionTerminatedEvent
            | ModelServiceStatusEvent,
        ) -> None:
            task_message = {"event": event.name, "session_id": str(event.session_id)}
            match event:
                case ModelServiceStatusEvent():
                    task_message["is_healthy"] = event.new_status.value
            await reporter.update(message=dump_json_str(task_message))

            match event:
                case SessionTerminatedEvent() | SessionCancelledEvent():
                    terminated_event.set()
                case ModelServiceStatusEvent():
                    async with root_ctx.db.begin_readonly_session() as db_sess:
                        session = await SessionRow.get_session(
                            db_sess,
                            result["sessionId"],
                            None,
                            kernel_loading_strategy=KernelLoadingStrategy.ALL_KERNELS,
                        )
                        await root_ctx.registry.destroy_session(
                            session,
                            forced=True,
                        )

        session_event_matcher = lambda args: args[0] == str(result["sessionId"])
        model_service_event_matcher = lambda args: args[1] == str(result["sessionId"])

        handlers: list[EventHandler] = [
            root_ctx.event_dispatcher.subscribe(
                SessionPreparingEvent,
                None,
                _handle_event,
                args_matcher=session_event_matcher,
            ),
            root_ctx.event_dispatcher.subscribe(
                SessionStartedEvent,
                None,
                _handle_event,
                args_matcher=session_event_matcher,
            ),
            root_ctx.event_dispatcher.subscribe(
                SessionCancelledEvent,
                None,
                _handle_event,
                args_matcher=session_event_matcher,
            ),
            root_ctx.event_dispatcher.subscribe(
                SessionTerminatedEvent,
                None,
                _handle_event,
                args_matcher=session_event_matcher,
            ),
            root_ctx.event_dispatcher.subscribe(
                ModelServiceStatusEvent,
                None,
                _handle_event,
                args_matcher=model_service_event_matcher,
            ),
        ]

        try:
            await terminated_event.wait()
        finally:
            for handler in handlers:
                root_ctx.event_dispatcher.unsubscribe(handler)

    task_id = await background_task_manager.start(_task)
    return TryStartResponseModel(task_id=str(task_id))


@auth_required
@server_status_required(READ_ALLOWED)
@pydantic_response_api_handler
async def delete(request: web.Request) -> SuccessResponseModel:
    """
    Removes model service (and inference sessions for the service also).
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info(
        "SERVE.DELETE (email:{}, ak:{}, s:{})", request["user"]["email"], access_key, service_id
    )

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound
    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})

    async with root_ctx.db.begin_session() as db_sess:
        if len(endpoint.routings) == 0:
            query = (
                sa.update(EndpointRow)
                .where(EndpointRow.id == service_id)
                .values({"lifecycle_stage": EndpointLifecycle.DESTROYED})
            )
        else:
            query = (
                sa.update(EndpointRow)
                .where(EndpointRow.id == service_id)
                .values({
                    "replicas": 0,
                    "lifecycle_stage": EndpointLifecycle.DESTROYING,
                })
            )
        await db_sess.execute(query)
    return SuccessResponseModel()


@auth_required
@server_status_required(READ_ALLOWED)
@pydantic_response_api_handler
async def sync(request: web.Request) -> SuccessResponseModel:
    """
    Force syncs up-to-date model service information with AppProxy.
    In normal situations this will be automatically handled by Backend.AI schedulers,
    but this API is left open in case of unexpected restart of AppProxy process.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info("SERVE.SYNC (email:{}, ak:{}, s:{})", request["user"]["email"], access_key, service_id)

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound
    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})

    async with root_ctx.db.begin_session() as db_sess:
        await root_ctx.registry.update_appproxy_endpoint_routes(
            db_sess, endpoint, [r for r in endpoint.routings if r.status == RouteStatus.HEALTHY]
        )
    return SuccessResponseModel()


class ScaleRequestModel(LegacyBaseRequestModel):
    to: int = Field(description="Ideal number of inference sessions")


class ScaleResponseModel(LegacyBaseResponseModel):
    current_route_count: int
    target_count: int


@auth_required
@server_status_required(READ_ALLOWED)
@pydantic_params_api_handler(ScaleRequestModel)
async def scale(request: web.Request, params: ScaleRequestModel) -> ScaleResponseModel:
    """
    Updates ideal inference session count manually. Based on the difference of this number,
    inference sessions will be created or removed automatically.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info(
        "SERVE.SCALE (email:{}, ak:{}, s:{})", request["user"]["email"], access_key, service_id
    )

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound
    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})

    if params.to < 0:
        raise InvalidAPIParameters("Amount of desired session count cannot be a negative number")
    elif params.to > (
        _m := request["user"]["resource_policy"]["max_session_count_per_model_session"]
    ):
        raise InvalidAPIParameters(f"Cannot spawn more than {_m} sessions for a single service")

    async with root_ctx.db.begin_session() as db_sess:
        query = (
            sa.update(EndpointRow)
            .where(EndpointRow.id == service_id)
            .values({"replicas": params.to})
        )
        await db_sess.execute(query)
        return ScaleResponseModel(
            current_route_count=len(endpoint.routings), target_count=params.to
        )


class UpdateRouteRequestModel(LegacyBaseRequestModel):
    traffic_ratio: NonNegativeFloat


@auth_required
@server_status_required(READ_ALLOWED)
@pydantic_params_api_handler(UpdateRouteRequestModel)
async def update_route(
    request: web.Request, params: UpdateRouteRequestModel
) -> SuccessResponseModel:
    """
    Updates traffic bias of specific route.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])
    route_id = uuid.UUID(request.match_info["route_id"])

    log.info(
        "SERVE.UPDATE_ROUTE (email:{}, ak:{}, s:{}, r:{})",
        request["user"]["email"],
        access_key,
        service_id,
        route_id,
    )

    async with root_ctx.db.begin_session() as db_sess:
        try:
            route = await RoutingRow.get(db_sess, route_id, load_endpoint=True)
        except NoResultFound:
            raise ObjectNotFound
        if route.endpoint != service_id:
            raise ObjectNotFound
        await get_user_uuid_scopes(request, {"owner_uuid": route.endpoint_row.session_owner})

        query = (
            sa.update(RoutingRow)
            .where(RoutingRow.id == route_id)
            .values({"traffic_ratio": params.traffic_ratio})
        )
        await db_sess.execute(query)
        endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        try:
            await root_ctx.registry.update_appproxy_endpoint_routes(
                db_sess, endpoint, [r for r in endpoint.routes if r.status == RouteStatus.HEALTHY]
            )
        except aiohttp.ClientError as e:
            log.warning("failed to communicate with AppProxy endpoint: {}", str(e))
        return SuccessResponseModel()


@auth_required
@server_status_required(READ_ALLOWED)
@pydantic_response_api_handler
async def delete_route(request: web.Request) -> SuccessResponseModel:
    """
    Scales down the service by removing specific inference session.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])
    route_id = uuid.UUID(request.match_info["route_id"])

    log.info(
        "SERVE.DELETE_ROUTE (email:{}, ak:{}, s:{})",
        request["user"]["email"],
        access_key,
        service_id,
    )
    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            route = await RoutingRow.get(db_sess, route_id, load_endpoint=True, load_session=True)
        except NoResultFound:
            raise ObjectNotFound
        if route.endpoint != service_id:
            raise ObjectNotFound
    await get_user_uuid_scopes(request, {"owner_uuid": route.endpoint_row.session_owner})
    if route.status == RouteStatus.PROVISIONING:
        raise InvalidAPIParameters("Cannot remove route in PROVISIONING status")

    await root_ctx.registry.destroy_session(
        route.session_row,
        forced=False,
        reason=KernelLifecycleEventReason.SERVICE_SCALED_DOWN,
    )

    async with root_ctx.db.begin_session() as db_sess:
        query = (
            sa.update(EndpointRow)
            .where(EndpointRow.id == service_id)
            .values({"replicas": route.endpoint_row.replicas - 1})
        )
        await db_sess.execute(query)
        return SuccessResponseModel()


class TokenRequestModel(LegacyBaseRequestModel):
    duration: tv.TimeDuration | None = Field(
        default=None, description="The lifetime duration of the token."
    )
    valid_until: int | None = Field(
        default=None,
        description="The absolute token expiry date expressed in the Unix epoch format.",
    )
    expires_at: int = Field(
        default=-1,
        description="The expiration timestamp computed from duration or valid_until.",
    )

    @model_validator(mode="after")
    def check_lifetime(self) -> Self:
        now = datetime.now()
        if self.valid_until is not None:
            self.expires_at = self.valid_until
        elif self.duration is not None:
            self.expires_at = int((now + self.duration).timestamp())
        else:
            raise ValueError("Either valid_until or duration must be specified.")
        if now.timestamp() > self.expires_at:
            raise ValueError("The expiration time cannot be in the past.")
        return self


class TokenResponseModel(LegacyBaseResponseModel):
    token: str


@auth_required
@server_status_required(READ_ALLOWED)
@pydantic_params_api_handler(TokenRequestModel)
async def generate_token(request: web.Request, params: TokenRequestModel) -> TokenResponseModel:
    """
    Generates a token which acts as an API key to authenticate when calling model service endpoint.
    If both duration and valid_until is not set then the AppProxy will determine appropriate lifetime of the token.
    duration and valid_until can't be both specified.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info(
        "SERVE.GENERATE_TOKEN (email:{}, ak:{}, s:{})",
        request["user"]["email"],
        access_key,
        service_id,
    )

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound
        query = (
            sa.select([scaling_groups.c.wsproxy_addr, scaling_groups.c.wsproxy_api_token])
            .select_from(scaling_groups)
            .where((scaling_groups.c.name == endpoint.resource_group))
        )

        result = await db_sess.execute(query)
        sgroup = result.first()
        wsproxy_addr = sgroup["wsproxy_addr"]
        wsproxy_api_token = sgroup["wsproxy_api_token"]

    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})

    body = {"user_uuid": str(endpoint.session_owner), "exp": params.expires_at}
    async with aiohttp.ClientSession() as session:
        async with session.post(
            f"{wsproxy_addr}/v2/endpoints/{endpoint.id}/token",
            json=body,
            headers={
                "X-BackendAI-Token": wsproxy_api_token,
            },
        ) as resp:
            token_json = await resp.json()
            token = token_json["token"]

    async with root_ctx.db.begin_session() as db_sess:
        token_row = EndpointTokenRow(
            uuid.uuid4(),
            token,
            endpoint.id,
            endpoint.domain,
            endpoint.project,
            endpoint.session_owner,
        )
        db_sess.add(token_row)
        await db_sess.commit()
        return TokenResponseModel(token=token)


class ErrorInfoModel(BaseFieldModel):
    session_id: uuid.UUID | None
    error: dict[str, Any]


class ErrorListResponseModel(LegacyBaseResponseModel):
    errors: list[ErrorInfoModel]
    retries: int


@auth_required
@server_status_required(READ_ALLOWED)
@pydantic_response_api_handler
async def list_errors(request: web.Request) -> ErrorListResponseModel:
    """
    List errors raised while trying to create the inference sessions. Backend.AI will
    stop trying to create an inference session for the model service if six (6) error stacks
    up. The only way to clear the error and retry spawning session is to call
    `clear_error` (POST /services/{service_id}/errors/clear) API.
    """
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info(
        "SERVE.LIST_ERRORS (email:{}, ak:{}, s:{})",
        request["user"]["email"],
        access_key,
        service_id,
    )

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound
    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})

    error_routes = [r for r in endpoint.routings if r.status == RouteStatus.FAILED_TO_START]

    return ErrorListResponseModel(
        errors=[
            ErrorInfoModel(
                session_id=route.error_data.get("session_id"), error=route.error_data["errors"]
            )
            for route in error_routes
        ],
        retries=endpoint.retries,
    )


@auth_required
@server_status_required(READ_ALLOWED)
async def clear_error(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info(
        "SERVE.CLEAR_ERROR (email:{}, ak:{}, s:{})",
        request["user"]["email"],
        access_key,
        service_id,
    )

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound
    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})

    async with root_ctx.db.begin_session() as db_sess:
        query = sa.delete(RoutingRow).where(
            (RoutingRow.endpoint == service_id) & (RoutingRow.status == RouteStatus.FAILED_TO_START)
        )
        await db_sess.execute(query)
        query = sa.update(EndpointRow).values({"retries": 0}).where(EndpointRow.id == endpoint.id)
        await db_sess.execute(query)

    return web.Response(status=HTTPStatus.NO_CONTENT)


class RuntimeInfo(BaseFieldModel):
    name: str = Field(description="Identifier to be passed later inside request body")
    human_readable_name: str = Field(description="Use this value as displayed label to user")


class RuntimeInfoModel(LegacyBaseResponseModel):
    runtimes: list[RuntimeInfo]


@auth_required
@server_status_required(READ_ALLOWED)
@pydantic_response_api_handler
async def list_supported_runtimes(request: web.Request) -> RuntimeInfoModel:
    return RuntimeInfoModel(
        runtimes=[
            RuntimeInfo(name=v.value, human_readable_name=MODEL_SERVICE_RUNTIME_PROFILES[v].name)
            for v in RuntimeVariant
        ]
    )


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    database_ptask_group: aiotools.PersistentTaskGroup


async def init(app: web.Application) -> None:
    app_ctx: PrivateContext = app["services.context"]
    app_ctx.database_ptask_group = aiotools.PersistentTaskGroup()


async def shutdown(app: web.Application) -> None:
    app_ctx: PrivateContext = app["services.context"]
    await app_ctx.database_ptask_group.shutdown()


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "services"
    app["api_versions"] = (4, 5)
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app["services.context"] = PrivateContext()
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    root_resource = cors.add(app.router.add_resource(r""))
    cors.add(root_resource.add_route("GET", list_serve))
    cors.add(root_resource.add_route("POST", create))
    cors.add(add_route("POST", "/_/try", try_start))
    cors.add(add_route("GET", "/_/runtimes", list_supported_runtimes))
    cors.add(add_route("GET", "/{service_id}", get_info))
    cors.add(add_route("DELETE", "/{service_id}", delete))
    cors.add(add_route("GET", "/{service_id}/errors", list_errors))
    cors.add(add_route("POST", "/{service_id}/errors/clear", clear_error))
    cors.add(add_route("POST", "/{service_id}/scale", scale))
    cors.add(add_route("POST", "/{service_id}/sync", sync))
    cors.add(add_route("PUT", "/{service_id}/routings/{route_id}", update_route))
    cors.add(add_route("DELETE", "/{service_id}/routings/{route_id}", delete_route))
    cors.add(add_route("POST", "/{service_id}/token", generate_token))
    return app, []



from dataclasses import dataclass
from typing import Any, Optional, override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.group.actions.base import GroupAction


# TODO: Change to batch action
@dataclass
class UsagePerMonthAction(GroupAction):
    month: str
    group_ids: Optional[list[UUID]] = None

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    def operation_type(self):
        return "usage_per_month"


@dataclass
class UsagePerMonthActionResult(BaseActionResult):
    # TODO: Define return type
    result: list[Any]

    @override
    def entity_id(self) -> Optional[str]:
        return None



import uuid
from pathlib import PurePosixPath
from typing import (
    Any,
    Optional,
    cast,
)

import aiohttp
import sqlalchemy as sa
from aiohttp import hdrs, web
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload

from ai.backend.common.bgtask import BackgroundTaskManager
from ai.backend.common.defs import VFOLDER_GROUP_PERMISSION_MODE
from ai.backend.common.types import (
    QuotaScopeID,
    QuotaScopeType,
    VFolderHostPermission,
    VFolderID,
    VFolderUsageMode,
)
from ai.backend.manager.api.exceptions import ObjectNotFound, StorageProxyError
from ai.backend.manager.config import DEFAULT_CHUNK_SIZE, SharedConfig
from ai.backend.manager.models.endpoint import EndpointLifecycle, EndpointRow
from ai.backend.manager.models.group import GroupRow, ProjectType
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.user import UserRole, UserRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine, execute_with_txn_retry
from ai.backend.manager.models.vfolder import (
    HARD_DELETED_VFOLDER_STATUSES,
    VFolderCloneInfo,
    VFolderDeletionInfo,
    VFolderOperationStatus,
    VFolderOwnershipType,
    VFolderPermission,
    VFolderPermissionRow,
    VFolderRow,
    VFolderStatusSet,
    delete_vfolder_relation_rows,
    ensure_host_permission_allowed,
    filter_host_allowed_permission,
    initiate_vfolder_clone,
    initiate_vfolder_deletion,
    is_unmanaged,
    query_accessible_vfolders,
    update_vfolder_status,
    verify_vfolder_name,
    vfolder_status_map,
    vfolders,
)

# from ai.backend.manager.types import SENTINEL
from ..actions.base import (
    CloneVFolderAction,
    CloneVFolderActionResult,
    CreateVFolderAction,
    CreateVFolderActionResult,
    DeleteForeverVFolderAction,
    DeleteForeverVFolderActionResult,
    ForceDeleteVFolderAction,
    ForceDeleteVFolderActionResult,
    GetTaskLogsAction,
    GetTaskLogsActionResult,
    GetVFolderAction,
    GetVFolderActionResult,
    ListVFolderAction,
    ListVFolderActionResult,
    MoveToTrashVFolderAction,
    MoveToTrashVFolderActionResult,
    RestoreVFolderFromTrashAction,
    RestoreVFolderFromTrashActionResult,
    UpdateVFolderAttributeAction,
    UpdateVFolderAttributeActionResult,
)
from ..exceptions import (
    Forbidden,
    InvalidParameter,
    ModelServiceDependencyNotCleared,
    ProjectNotFound,
    TooManyVFoldersFound,
    VFolderAlreadyExists,
    VFolderCreationFailure,
    VFolderFilterStatusFailed,
    VFolderFilterStatusNotAvailable,
    VFolderNotFound,
)
from ..types import VFolderBaseInfo, VFolderOwnershipInfo, VFolderUsageInfo


async def _check_vfolder_status(
    vfolder_status: VFolderOperationStatus,
    status: VFolderStatusSet,
) -> None:
    """
    Checks if the target vfolder status matches one of the status sets aliased by `status` VFolderStatusSet,
    """

    available_vf_statuses = vfolder_status_map.get(status)
    if not available_vf_statuses:
        raise VFolderFilterStatusNotAvailable
    if vfolder_status not in available_vf_statuses:
        raise VFolderFilterStatusFailed


class VFolderService:
    _db: ExtendedAsyncSAEngine
    _shared_config: SharedConfig
    _storage_manager: StorageSessionManager
    _background_task_manager: BackgroundTaskManager

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        shared_config: SharedConfig,
        storage_manager: StorageSessionManager,
        background_task_manager: BackgroundTaskManager,
    ) -> None:
        self._db = db
        self._shared_config = shared_config
        self._storage_manager = storage_manager
        self._background_task_manager = background_task_manager

    async def create(self, action: CreateVFolderAction) -> CreateVFolderActionResult:
        user_role = action.user_role
        user_uuid = action.user_uuid
        keypair_resource_policy = action.keypair_resource_policy
        domain_name = action.domain_name
        group_id_or_name = action.group_id_or_name
        folder_host = action.folder_host
        unmanaged_path = action.unmanaged_path
        # Resolve host for the new virtual folder.
        if not folder_host:
            folder_host = await self._shared_config.etcd.get("volumes/default_host")
            if not folder_host:
                raise InvalidParameter(
                    "You must specify the vfolder host because the default host is not configured."
                )
        # Check if user is trying to created unmanaged vFolder
        if unmanaged_path:
            # Approve only if user is Admin or Superadmin
            if user_role not in (UserRole.ADMIN, UserRole.SUPERADMIN):
                raise Forbidden("Insufficient permission")
                # Assign ghost host to unmanaged vfolder

        allowed_vfolder_types = await self._shared_config.get_vfolder_types()

        if action.name.startswith(".") and action.name != ".local":
            if action.group_id_or_name is not None:
                raise InvalidParameter("dot-prefixed vfolders cannot be a group folder.")

        group_uuid: Optional[uuid.UUID] = None
        group_type: Optional[ProjectType] = None
        max_vfolder_count: int
        max_quota_scope_size: int
        container_uid: Optional[int] = None

        async with self._db.begin_session() as sess:
            match group_id_or_name:
                case str():
                    # Convert the group name to group uuid.
                    # log.debug("group_id_or_name(str):{}", group_id_or_name)
                    query = (
                        sa.select(GroupRow)
                        .where(
                            (GroupRow.domain_name == domain_name)
                            & (GroupRow.name == group_id_or_name)
                        )
                        .options(selectinload(GroupRow.resource_policy_row))
                    )
                    result = await sess.execute(query)
                    group_row = cast(Optional[GroupRow], result.scalar())
                    if group_row is None:
                        raise ProjectNotFound(group_id_or_name)
                    _gid, max_vfolder_count, max_quota_scope_size = (
                        cast(Optional[uuid.UUID], group_row.id),
                        cast(int, group_row.resource_policy_row.max_vfolder_count),
                        cast(int, group_row.resource_policy_row.max_quota_scope_size),
                    )
                    if _gid is None:
                        raise ProjectNotFound(group_id_or_name)
                    group_uuid = _gid
                    group_type = cast(ProjectType, group_row.type)
                case uuid.UUID():
                    # Check if the group belongs to the current domain.
                    # log.debug("group_id_or_name(uuid):{}", group_id_or_name)
                    query = (
                        sa.select(GroupRow)
                        .where(
                            (GroupRow.domain_name == domain_name)
                            & (GroupRow.id == group_id_or_name)
                        )
                        .options(selectinload(GroupRow.resource_policy_row))
                    )
                    result = await sess.execute(query)
                    group_row = cast(Optional[GroupRow], result.scalar())
                    if group_row is None:
                        raise ProjectNotFound(group_id_or_name)
                    _gid, max_vfolder_count, max_quota_scope_size = (
                        group_row.id,
                        cast(int, group_row.resource_policy_row.max_vfolder_count),
                        cast(int, group_row.resource_policy_row.max_quota_scope_size),
                    )
                    if _gid is None:
                        raise ProjectNotFound(group_id_or_name)
                    group_uuid = group_id_or_name
                    group_type = cast(ProjectType, group_row.type)
                case None:
                    query = (
                        sa.select(UserRow)
                        .where(UserRow.uuid == user_uuid)
                        .options(selectinload(UserRow.resource_policy_row))
                    )
                    result = await sess.execute(query)
                    user_row = result.scalar()
                    max_vfolder_count, max_quota_scope_size = (
                        cast(int, user_row.resource_policy_row.max_vfolder_count),
                        cast(int, user_row.resource_policy_row.max_quota_scope_size),
                    )
                    container_uid = cast(Optional[int], user_row.container_uid)
                case _:
                    raise ProjectNotFound(group_id_or_name)

            vfolder_permission_mode = (
                VFOLDER_GROUP_PERMISSION_MODE if container_uid is not None else None
            )

            # Check if group exists when it's given a non-empty value.
            if group_id_or_name and group_uuid is None:
                raise ProjectNotFound(group_id_or_name)

            # Determine the ownership type and the quota scope ID.
            if group_uuid is not None:
                ownership_type = "group"
                quota_scope_id = QuotaScopeID(QuotaScopeType.PROJECT, group_uuid)
                if (
                    user_role not in (UserRole.SUPERADMIN, UserRole.ADMIN)
                    and group_type != ProjectType.MODEL_STORE
                ):
                    raise Forbidden("no permission")
            else:
                ownership_type = "user"
                quota_scope_id = QuotaScopeID(QuotaScopeType.USER, user_uuid)
            if ownership_type not in allowed_vfolder_types:
                raise InvalidParameter(
                    f"{ownership_type}-owned vfolder is not allowed in this cluster"
                )

        if group_type == ProjectType.MODEL_STORE:
            if action.mount_permission != VFolderPermission.READ_WRITE:
                raise InvalidParameter(
                    "Setting custom permission is not supported for model store vfolder"
                )
            if action.usage_mode != VFolderUsageMode.MODEL:
                raise InvalidParameter(
                    "Only Model VFolder can be created under the model store project"
                )

        async with self._db.begin() as conn:
            await ensure_host_permission_allowed(
                conn,
                folder_host,
                allowed_vfolder_types=allowed_vfolder_types,
                user_uuid=user_uuid,
                resource_policy=keypair_resource_policy,
                domain_name=domain_name,
                group_id=group_uuid,
                permission=VFolderHostPermission.CREATE,
            )

            # Check resource policy's max_vfolder_count
            if max_vfolder_count > 0:
                if ownership_type == "user":
                    query = (
                        sa.select(sa.func.count())
                        .select_from(VFolderRow)
                        .where(
                            (VFolderRow.user == user_uuid)
                            & (VFolderRow.status.not_in(HARD_DELETED_VFOLDER_STATUSES))
                        )
                    )
                else:
                    assert group_uuid is not None
                    query = (
                        sa.select(sa.func.count())
                        .select_from(VFolderRow)
                        .where(
                            (VFolderRow.group == group_uuid)
                            & (VFolderRow.status.not_in(HARD_DELETED_VFOLDER_STATUSES))
                        )
                    )
                result = cast(int, await conn.scalar(query))
                if result >= max_vfolder_count:
                    raise InvalidParameter("You cannot create more vfolders.")

            # DEPRECATED: Limit vfolder size quota if it is larger than max_vfolder_size of the resource policy.
            # max_vfolder_size = resource_policy.get("max_vfolder_size", 0)
            # if max_vfolder_size > 0 and (
            #     params["quota"] is None or params["quota"] <= 0 or params["quota"] > max_vfolder_size
            # ):
            #     params["quota"] = max_vfolder_size

            # Prevent creation of vfolder with duplicated name on all hosts.
            extra_vf_conds = [
                (VFolderRow.name == action.name),
                (VFolderRow.status.not_in(HARD_DELETED_VFOLDER_STATUSES)),
            ]
            entries = await query_accessible_vfolders(
                conn,
                user_uuid,
                user_role=user_role,
                domain_name=domain_name,
                allowed_vfolder_types=allowed_vfolder_types,
                extra_vf_conds=(sa.and_(*extra_vf_conds)),
            )
            if len(entries) > 0:
                raise VFolderAlreadyExists(
                    f"VFolder with the given name already exists. ({action.name})"
                )
            try:
                folder_id = uuid.uuid4()
                vfid = VFolderID(quota_scope_id, folder_id)
                if not unmanaged_path:
                    options = {}
                    if max_quota_scope_size and max_quota_scope_size > 0:
                        options["initial_max_size_for_quota_scope"] = max_quota_scope_size
                    body_data: dict[str, Any] = {
                        "volume": self._storage_manager.get_proxy_and_volume(
                            folder_host, is_unmanaged(unmanaged_path)
                        )[1],
                        "vfid": str(vfid),
                        "options": options,
                    }
                    if vfolder_permission_mode is not None:
                        body_data["mode"] = vfolder_permission_mode
                    async with self._storage_manager.request(
                        folder_host,
                        "POST",
                        "folder/create",
                        json=body_data,
                    ):
                        pass
            except aiohttp.ClientResponseError as e:
                raise VFolderCreationFailure from e

            # By default model store VFolder should be considered as read only for every users but without the creator
            if group_type == ProjectType.MODEL_STORE:
                action.mount_permission = VFolderPermission.READ_ONLY

            # TODO: include quota scope ID in the database
            # TODO: include quota scope ID in the API response
            insert_values = {
                "id": vfid.folder_id.hex,
                "name": action.name,
                "domain_name": domain_name,
                "quota_scope_id": str(quota_scope_id),
                "usage_mode": action.usage_mode,
                "permission": action.mount_permission,
                "last_used": None,
                "host": folder_host,
                "creator": action.creator_email,
                "ownership_type": VFolderOwnershipType(ownership_type),
                "user": user_uuid if ownership_type == "user" else None,
                "group": group_uuid if ownership_type == "group" else None,
                "unmanaged_path": unmanaged_path,
                "cloneable": action.cloneable,
                "status": VFolderOperationStatus.READY,
            }
            try:
                query = sa.insert(VFolderRow, insert_values)
                result = await conn.execute(query)

                # Here we grant creator the permission to alter VFolder contents
                if (
                    group_type == ProjectType.MODEL_STORE
                    and ownership_type == VFolderOwnershipType.USER
                ):
                    query = sa.insert(VFolderPermissionRow).values({
                        "user": user_uuid,
                        "vfolder": vfid.folder_id.hex,
                        "permission": VFolderPermission.OWNER_PERM,
                    })
                    await conn.execute(query)
            except sa.exc.DataError:
                raise InvalidParameter
            assert result.rowcount == 1

        return CreateVFolderActionResult(
            id=vfid.folder_id,
            name=action.name,
            quota_scope_id=quota_scope_id,
            host=folder_host,
            unmanaged_path=unmanaged_path,
            usage_mode=action.usage_mode,
            mount_permission=action.mount_permission,
            user_uuid=user_uuid if ownership_type == "user" else None,
            group_uuid=group_uuid if ownership_type == "group" else None,
            creator_email=action.creator_email,
            ownership_type=VFolderOwnershipType(ownership_type),
            cloneable=action.cloneable,
            status=VFolderOperationStatus.READY,
        )

    async def update_attribute(
        self, action: UpdateVFolderAttributeAction
    ) -> UpdateVFolderAttributeActionResult:
        modifier = action.modifier
        allowed_vfolder_types = await self._shared_config.get_vfolder_types()

        async def _update(db_session: AsyncSession) -> None:
            requester_user_row = await db_session.scalar(
                sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
            )
            requester_user_row = cast(UserRow, requester_user_row)
            vfolder_dicts = await query_accessible_vfolders(
                db_session.bind,
                action.user_uuid,
                allow_privileged_access=True,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
            )
            if not vfolder_dicts:
                raise VFolderNotFound
            query_vfolder = sa.select(VFolderRow).where(VFolderRow.id == action.vfolder_uuid)
            vfolder_row = await db_session.scalar(query_vfolder)
            vfolder_row = cast(VFolderRow, vfolder_row)
            try:
                new_name = modifier.name.value()
            except ValueError:
                pass
            else:
                for row in vfolder_dicts:
                    if row["name"] == new_name:
                        raise InvalidParameter(
                            "One of your accessible vfolders already has the name you requested."
                        )
            to_update = modifier.fields_to_update()
            for key, value in to_update.items():
                setattr(vfolder_row, key, value)

        async with self._db.connect() as db_conn:
            await execute_with_txn_retry(_update, self._db.begin_session, db_conn)
        return UpdateVFolderAttributeActionResult(vfolder_uuid=action.vfolder_uuid)

    async def get(self, action: GetVFolderAction) -> GetVFolderActionResult:
        allowed_vfolder_types = await self._shared_config.get_vfolder_types()
        async with self._db.begin_session() as db_session:
            requester_user_row = await db_session.scalar(
                sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
            )
            requester_user_row = cast(UserRow, requester_user_row)
            vfolder_dicts = await query_accessible_vfolders(
                db_session.bind,
                action.user_uuid,
                allow_privileged_access=True,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
                extra_vf_conds=(VFolderRow.id == action.vfolder_uuid),
            )
            vfolder_row = vfolder_dicts[0]
        if vfolder_row["permission"] is None:
            is_owner = True
            permission = VFolderPermission.OWNER_PERM
        else:
            is_owner = vfolder_row["is_owner"]
            permission = vfolder_row["permission"]
        proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(
            vfolder_row["host"], is_unmanaged(vfolder_row["unmanaged_path"])
        )
        async with self._storage_manager.request(
            proxy_name,
            "GET",
            "folder/usage",
            json={
                "volume": volume_name,
                "vfid": str(VFolderID(vfolder_row["quota_scope_id"], vfolder_row["id"])),
            },
        ) as (_, storage_resp):
            usage = await storage_resp.json()
            usage_info = VFolderUsageInfo(
                used_bytes=usage["used_bytes"],
                num_files=usage["file_count"],
            )
        return GetVFolderActionResult(
            user_uuid=action.user_uuid,
            base_info=VFolderBaseInfo(
                id=vfolder_row["id"],
                quota_scope_id=vfolder_row["quota_scope_id"],
                name=vfolder_row["name"],
                host=vfolder_row["host"],
                status=vfolder_row["status"],
                unmanaged_path=vfolder_row["unmanaged_path"],
                mount_permission=permission,
                usage_mode=vfolder_row["usage_mode"],
                created_at=vfolder_row["created_at"],
                cloneable=vfolder_row["cloneable"],
            ),
            ownership_info=VFolderOwnershipInfo(
                creator_email=vfolder_row["creator"],
                ownership_type=vfolder_row["ownership_type"],
                is_owner=is_owner,
                user_uuid=vfolder_row["user"],
                group_uuid=vfolder_row["group"],
            ),
            usage_info=usage_info,
        )

    async def list(self, action: ListVFolderAction) -> ListVFolderActionResult:
        allowed_vfolder_types = await self._shared_config.get_vfolder_types()
        async with self._db.begin_session() as db_session:
            requester_user_row = await db_session.scalar(
                sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
            )
            requester_user_row = cast(UserRow, requester_user_row)
            vfolder_dicts = await query_accessible_vfolders(
                db_session.bind,
                action.user_uuid,
                allow_privileged_access=False,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
            )
        vfolders = [
            (
                VFolderBaseInfo(
                    id=entry["id"],
                    quota_scope_id=entry["quota_scope_id"],
                    name=entry["name"],
                    host=entry["host"],
                    status=entry["status"],
                    unmanaged_path=entry["unmanaged_path"],
                    mount_permission=entry["permission"],
                    usage_mode=entry["usage_mode"],
                    created_at=entry["created_at"],
                    cloneable=entry["cloneable"],
                ),
                VFolderOwnershipInfo(
                    creator_email=entry["creator"],
                    ownership_type=entry["ownership_type"],
                    is_owner=entry["is_owner"],
                    user_uuid=entry["user"],
                    group_uuid=entry["group"],
                ),
            )
            for entry in vfolder_dicts
        ]

        return ListVFolderActionResult(
            user_uuid=action.user_uuid,
            vfolders=vfolders,
        )

    async def move_to_trash(
        self, action: MoveToTrashVFolderAction
    ) -> MoveToTrashVFolderActionResult:
        # Only the effective folder owner can delete the folder.
        allowed_vfolder_types = await self._shared_config.get_vfolder_types()
        async with self._db.connect() as db_conn:
            async with self._db.begin_session(db_conn) as db_session:
                requester_user_row = await db_session.scalar(
                    sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
                )
                requester_user_row = cast(UserRow, requester_user_row)
                vfolder_dicts = await query_accessible_vfolders(
                    db_session.bind,
                    action.user_uuid,
                    allow_privileged_access=True,
                    user_role=requester_user_row.role,
                    allowed_vfolder_types=allowed_vfolder_types,
                    domain_name=requester_user_row.domain_name,
                    extra_vf_conds=(VFolderRow.id == action.vfolder_uuid),
                )
                vfolder_row = vfolder_dicts[0]
                if not vfolder_row["is_owner"]:
                    raise InvalidParameter("Cannot delete the vfolder that is not owned by myself.")
                await _check_vfolder_status(vfolder_row["status"], VFolderStatusSet.DELETABLE)
                # perform extra check to make sure records of alive model service not removed by foreign key rule
                if vfolder_row["usage_mode"] == VFolderUsageMode.MODEL:
                    live_endpoints = await EndpointRow.list_by_model(db_session, vfolder_row["id"])
                    if (
                        len([
                            e
                            for e in live_endpoints
                            if e.lifecycle_stage == EndpointLifecycle.CREATED
                        ])
                        > 0
                    ):
                        raise ModelServiceDependencyNotCleared
                folder_host = vfolder_row["host"]
                await ensure_host_permission_allowed(
                    db_session.bind,
                    folder_host,
                    allowed_vfolder_types=allowed_vfolder_types,
                    user_uuid=action.user_uuid,
                    resource_policy=action.keypair_resource_policy,
                    domain_name=requester_user_row.domain_name,
                    permission=VFolderHostPermission.DELETE,
                )

                vfolder_row_ids = (vfolder_row["id"],)
            await delete_vfolder_relation_rows(db_conn, self._db.begin_session, vfolder_row_ids)
        await update_vfolder_status(
            self._db,
            vfolder_row_ids,
            VFolderOperationStatus.DELETE_PENDING,
        )
        return MoveToTrashVFolderActionResult(vfolder_uuid=action.vfolder_uuid)

    async def restore(
        self, action: RestoreVFolderFromTrashAction
    ) -> RestoreVFolderFromTrashActionResult:
        allowed_vfolder_types = await self._shared_config.get_vfolder_types()

        async with self._db.begin_session() as db_session:
            requester_user_row = await db_session.scalar(
                sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
            )
            restore_targets = await query_accessible_vfolders(
                db_session.bind,
                action.user_uuid,
                allow_privileged_access=True,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
                extra_vf_conds=(VFolderRow.id == action.vfolder_uuid),
            )

            if len(restore_targets) > 1:
                raise TooManyVFoldersFound(restore_targets)
            elif len(restore_targets) == 0:
                raise InvalidParameter("No such vfolder.")

            row = restore_targets[0]
            await _check_vfolder_status(row["status"], VFolderStatusSet.RECOVERABLE)

        # Folder owner OR user who have DELETE permission can restore folder.
        if not row["is_owner"] and row["permission"] != VFolderPermission.RW_DELETE:
            raise InvalidParameter("Cannot restore the vfolder that is not owned by myself.")

        # fs-level mv may fail or take longer time
        # but let's complete the db transaction to reflect that it's deleted.
        await update_vfolder_status(self._db, (row["id"],), VFolderOperationStatus.READY)
        return RestoreVFolderFromTrashActionResult(vfolder_uuid=action.vfolder_uuid)

    async def delete_forever(
        self, action: DeleteForeverVFolderAction
    ) -> DeleteForeverVFolderActionResult:
        allowed_vfolder_types = await self._shared_config.get_vfolder_types()

        async with self._db.begin_session() as db_session:
            requester_user_row = await db_session.scalar(
                sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
            )
            entries = await query_accessible_vfolders(
                db_session.bind,
                action.user_uuid,
                allow_privileged_access=True,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
                extra_vf_conds=(VFolderRow.id == action.vfolder_uuid),
            )

            if len(entries) > 1:
                raise TooManyVFoldersFound(entries)
            elif len(entries) == 0:
                raise InvalidParameter("No such vfolder.")
            row = entries[0]
            await _check_vfolder_status(row["status"], VFolderStatusSet.PURGABLE)

        # fs-level deletion may fail or take longer time
        await initiate_vfolder_deletion(
            self._db,
            [VFolderDeletionInfo(VFolderID.from_row(row), row["host"], row["unmanaged_path"])],
            self._storage_manager,
        )
        return DeleteForeverVFolderActionResult(vfolder_uuid=action.vfolder_uuid)

    async def force_delete(
        self, action: ForceDeleteVFolderAction
    ) -> ForceDeleteVFolderActionResult:
        allowed_vfolder_types = await self._shared_config.get_vfolder_types()

        async with self._db.begin_session() as db_session:
            requester_user_row = await db_session.scalar(
                sa.select(UserRow).where(UserRow.uuid == action.user_uuid)
            )
            entries = await query_accessible_vfolders(
                db_session.bind,
                action.user_uuid,
                allow_privileged_access=True,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
                extra_vf_conds=(VFolderRow.id == action.vfolder_uuid),
            )
            row = entries[0]
            try:
                await _check_vfolder_status(row["status"], VFolderStatusSet.PURGABLE)
            except VFolderFilterStatusFailed:
                await _check_vfolder_status(row["status"], VFolderStatusSet.DELETABLE)
        await initiate_vfolder_deletion(
            self._db,
            [VFolderDeletionInfo(VFolderID.from_row(row), row["host"], row["unmanaged_path"])],
            storage_manager=self._storage_manager,
            force=True,
        )
        return ForceDeleteVFolderActionResult(vfolder_uuid=action.vfolder_uuid)

    async def clone(self, action: CloneVFolderAction) -> CloneVFolderActionResult:
        allowed_vfolder_types = await self._shared_config.get_vfolder_types()
        if "user" not in allowed_vfolder_types:
            raise InvalidParameter("user vfolder cannot be created in this host")
        requester_user_row = await UserRow.get_by_id_with_policies(
            action.requester_user_uuid, db=self._db
        )
        if requester_user_row is None:
            raise InvalidParameter("No such user.")
        async with self._db.begin_readonly_session() as db_session:
            entries = await query_accessible_vfolders(
                db_session.bind,
                action.requester_user_uuid,
                allow_privileged_access=True,
                user_role=requester_user_row.role,
                allowed_vfolder_types=allowed_vfolder_types,
                domain_name=requester_user_row.domain_name,
            )
            for entry in entries:
                if entry["id"] == action.source_vfolder_uuid:
                    row = entry
                    break
            else:
                raise InvalidParameter("No such vfolder.")
        project_row: Optional[GroupRow] = None
        if (project_id := row["group"]) is not None:
            project_row = await GroupRow.get_by_id_with_policies(project_id, db=self._db)
        async with self._db.begin_session() as db_session:
            domain_name = requester_user_row.domain_name
            source_folder_host = row["host"]
            source_folder_id = VFolderID(row["quota_scope_id"], row["id"])
            target_folder_host = action.target_host or source_folder_host
            target_quota_scope_id = "..."  # TODO: implement
            source_unmanaged_path = row["unmanaged_path"]
            source_proxy_name, source_volume_name = self._storage_manager.get_proxy_and_volume(
                source_folder_host, is_unmanaged(source_unmanaged_path)
            )
            target_proxy_name, target_volume_name = self._storage_manager.get_proxy_and_volume(
                target_folder_host,
            )

            # check if the source vfolder is allowed to be cloned
            if not row["cloneable"]:
                raise Forbidden("The source vfolder is not permitted to be cloned.")

            if action.target_name.startswith("."):
                for entry in entries:
                    if entry["name"] == action.target_name:
                        raise VFolderAlreadyExists
                    if entry["name"].startswith(".") and entry["path"] == action.target_name:
                        raise InvalidParameter("VFolder name conflicts with your dotfile.")

            if not target_folder_host:
                target_folder_host = await self._shared_config.etcd.get("volumes/default_host")
                if not target_folder_host:
                    raise InvalidParameter(
                        "You must specify the vfolder host because the default host is not configured."
                    )

            if not verify_vfolder_name(action.target_name):
                raise InvalidParameter(f"{action.target_name} is reserved for internal operations.")

            if source_proxy_name != target_proxy_name:
                raise InvalidParameter("proxy name of source and target vfolders must be equal.")

            if project_row is not None:
                vfolder_hosts = project_row.allowed_vfolder_hosts
                max_vfolder_count = project_row.resource_policy_row.max_vfolder_count
            else:
                vfolder_hosts = (
                    requester_user_row.main_keypair.resource_policy_row.allowed_vfolder_hosts
                )
                max_vfolder_count = requester_user_row.resource_policy_row.max_vfolder_count

            allowed_hosts = await filter_host_allowed_permission(
                db_session.bind,
                allowed_vfolder_types=allowed_vfolder_types,
                user_uuid=requester_user_row.uuid,
                resource_policy={"allowed_vfolder_hosts": vfolder_hosts},
                domain_name=domain_name,
            )
            if (
                target_folder_host not in allowed_hosts
                or VFolderHostPermission.CREATE not in allowed_hosts[target_folder_host]
            ):
                raise InvalidParameter(
                    f"`{VFolderHostPermission.CREATE}` Not allowed in vfolder"
                    f" host(`{target_folder_host}`)"
                )
            # TODO: handle legacy host lists assuming that volume names don't overlap?
            if target_folder_host not in allowed_hosts:
                raise InvalidParameter("You are not allowed to use this vfolder host.")

            # Check resource policy's max_vfolder_count
            if max_vfolder_count > 0:
                query = sa.select(sa.func.count()).where(
                    sa.and_(
                        VFolderRow.user == requester_user_row.uuid,
                        VFolderRow.status.not_in(HARD_DELETED_VFOLDER_STATUSES),
                    )
                )
                result = await db_session.scalar(query)
                if result >= max_vfolder_count:
                    raise InvalidParameter("You cannot create more vfolders.")

        task_id, target_folder_id = await initiate_vfolder_clone(
            self._db,
            VFolderCloneInfo(
                source_folder_id,
                source_folder_host,
                source_unmanaged_path,
                domain_name,
                target_quota_scope_id,
                action.target_name,
                target_folder_host,
                action.usage_mode,
                action.mount_permission,
                requester_user_row.email,
                requester_user_row.uuid,
                action.cloneable,
            ),
            self._storage_manager,
            self._background_task_manager,
        )

        # Return the information about the destination vfolder.
        return CloneVFolderActionResult(
            vfolder_uuid=source_folder_id.folder_id,
            target_vfolder_id=target_folder_id,
            target_vfolder_name=action.target_name,
            target_vfolder_host=target_folder_host,
            usage_mode=action.usage_mode,
            mount_permission=action.mount_permission,
            creator_email=requester_user_row.email,
            ownership_type=VFolderOwnershipType.USER,
            owner_user_uuid=requester_user_row.uuid,
            owner_group_uuid=None,
            cloneable=action.cloneable,
            bgtask_id=task_id,
        )

    async def get_task_logs(self, action: GetTaskLogsAction) -> GetTaskLogsActionResult:
        user_uuid = action.user_id
        user_role = action.user_role
        domain_name = action.domain_name
        kernel_id_str = action.kernel_id.hex
        request = action.request

        async with self._db.begin_readonly() as conn:
            matched_vfolders = await query_accessible_vfolders(
                conn,
                user_uuid,
                user_role=user_role,
                domain_name=domain_name,
                allowed_vfolder_types=["user"],
                extra_vf_conds=(vfolders.c.name == ".logs"),
            )
            if not matched_vfolders:
                raise ObjectNotFound(
                    extra_data={"vfolder_name": ".logs"},
                    object_name="vfolder",
                )
            log_vfolder = matched_vfolders[0]

        _proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(
            log_vfolder["host"], is_unmanaged(log_vfolder["unmanaged_path"])
        )
        response = web.StreamResponse(status=200)
        response.headers[hdrs.CONTENT_TYPE] = "text/plain"
        prepared = False

        try:
            async with self._storage_manager.request(
                log_vfolder["host"],
                "POST",
                "folder/file/fetch",
                json={
                    "volume": volume_name,
                    "vfid": str(VFolderID.from_row(log_vfolder)),
                    "relpath": str(
                        PurePosixPath("task")
                        / kernel_id_str[:2]
                        / kernel_id_str[2:4]
                        / f"{kernel_id_str[4:]}.log",
                    ),
                },
                raise_for_status=True,
            ) as (_, storage_resp):
                while True:
                    chunk = await storage_resp.content.read(DEFAULT_CHUNK_SIZE)
                    if not chunk:
                        break
                    if not prepared:
                        await response.prepare(request)
                        prepared = True
                    await response.write(chunk)
        except aiohttp.ClientResponseError as e:
            raise StorageProxyError(status=e.status, extra_msg=e.message)
        finally:
            if prepared:
                await response.write_eof()
        return GetTaskLogsActionResult(response=response, vfolder_data=log_vfolder)



import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Optional

from ai.backend.manager.actions.types import OperationStatus


@dataclass
class StartedActionMessage:
    """Message indicating that an action has started."""

    action_id: uuid.UUID
    action_type: str
    entity_id: Optional[str | uuid.UUID]
    request_id: Optional[str]
    entity_type: str
    operation_type: str
    created_at: datetime


@dataclass
class FinishedActionMessage:
    """Message indicating that an action has finished."""

    action_id: uuid.UUID
    action_type: str
    entity_id: Optional[str | uuid.UUID]  # TODO: Make this required?
    request_id: Optional[str]
    entity_type: str
    operation_type: str
    status: OperationStatus
    description: str
    created_at: datetime
    ended_at: datetime
    duration: timedelta


class AbstractReporter(ABC):
    @abstractmethod
    async def report_started(self, message: StartedActionMessage) -> None:
        raise NotImplementedError

    @abstractmethod
    async def report_finished(self, message: FinishedActionMessage) -> None:
        raise NotImplementedError



import asyncio
import base64
import functools
import logging
import secrets
import uuid
from dataclasses import asdict, dataclass
from datetime import datetime
from typing import Any, Dict, Iterable, Mapping, MutableMapping, Optional, Union, cast
from urllib.parse import urlparse

import aiohttp
import aiotools
import multidict
import sqlalchemy as sa
import trafaret as t
from dateutil.tz import tzutc
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import load_only, noload, selectinload

from ai.backend.common.bgtask import BackgroundTaskManager, ProgressReporter
from ai.backend.common.docker import DEFAULT_KERNEL_FEATURE, ImageRef, KernelFeatures, LabelName
from ai.backend.common.events import (
    BgtaskCancelledEvent,
    BgtaskDoneEvent,
    BgtaskFailedEvent,
)
from ai.backend.common.exception import BackendError, InvalidAPIParameters, UnknownImageReference
from ai.backend.common.json import load_json
from ai.backend.common.plugin.monitor import ErrorPluginContext
from ai.backend.common.types import (
    AccessKey,
    ImageAlias,
    ImageRegistry,
    SessionId,
    SessionTypes,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.api.exceptions import (
    AppNotFound,
    GenericForbidden,
    InternalServerError,
    KernelNotReady,
    QuotaExceeded,
    ServiceUnavailable,
    SessionAlreadyExists,
    SessionNotFound,
    TaskTemplateNotFound,
    TooManySessionsMatched,
    UnknownImageReferenceError,
)
from ai.backend.manager.api.scaling_group import query_wsproxy_status
from ai.backend.manager.api.session import (
    CustomizedImageVisibilityScope,
    drop_undefined,
    find_dependency_sessions,
    find_dependent_sessions,
    overwritten_param_check,
)
from ai.backend.manager.api.utils import undefined
from ai.backend.manager.idle import IdleCheckerHost
from ai.backend.manager.models.container_registry import ContainerRegistryRow
from ai.backend.manager.models.group import GroupRow, groups
from ai.backend.manager.models.image import ImageIdentifier, ImageRow, rescan_images
from ai.backend.manager.models.kernel import (
    KernelRow,
)
from ai.backend.manager.models.scaling_group import scaling_groups
from ai.backend.manager.models.session import (
    DEAD_SESSION_STATUSES,
    PRIVATE_SESSION_TYPES,
    KernelLoadingStrategy,
    SessionRow,
    SessionStatus,
)
from ai.backend.manager.models.session_template import session_templates
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine, execute_with_txn_retry
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.services.session.actions.check_and_transit_status import (
    CheckAndTransitStatusAction,
    CheckAndTransitStatusActionResult,
    CheckAndTransitStatusBatchAction,
    CheckAndTransitStatusBatchActionResult,
)
from ai.backend.manager.services.session.actions.commit_session import (
    CommitSessionAction,
    CommitSessionActionResult,
)
from ai.backend.manager.services.session.actions.complete import (
    CompleteAction,
    CompleteActionResult,
)
from ai.backend.manager.services.session.actions.convert_session_to_image import (
    ConvertSessionToImageAction,
    ConvertSessionToImageActionResult,
)
from ai.backend.manager.services.session.actions.create_cluster import (
    CreateClusterAction,
    CreateClusterActionResult,
)
from ai.backend.manager.services.session.actions.create_from_params import (
    CreateFromParamsAction,
    CreateFromParamsActionResult,
)
from ai.backend.manager.services.session.actions.create_from_template import (
    CreateFromTemplateAction,
    CreateFromTemplateActionResult,
)
from ai.backend.manager.services.session.actions.destory_session import (
    DestroySessionAction,
    DestroySessionActionResult,
)
from ai.backend.manager.services.session.actions.download_file import (
    DownloadFileAction,
    DownloadFileActionResult,
)
from ai.backend.manager.services.session.actions.download_files import (
    DownloadFilesAction,
    DownloadFilesActionResult,
)
from ai.backend.manager.services.session.actions.execute_session import (
    ExecuteSessionAction,
    ExecuteSessionActionResult,
)
from ai.backend.manager.services.session.actions.get_abusing_report import (
    GetAbusingReportAction,
    GetAbusingReportActionResult,
)
from ai.backend.manager.services.session.actions.get_commit_status import (
    GetCommitStatusAction,
    GetCommitStatusActionResult,
)
from ai.backend.manager.services.session.actions.get_container_logs import (
    GetContainerLogsAction,
    GetContainerLogsActionResult,
)
from ai.backend.manager.services.session.actions.get_dependency_graph import (
    GetDependencyGraphAction,
    GetDependencyGraphActionResult,
)
from ai.backend.manager.services.session.actions.get_direct_access_info import (
    GetDirectAccessInfoAction,
    GetDirectAccessInfoActionResult,
)
from ai.backend.manager.services.session.actions.get_session_info import (
    GetSessionInfoAction,
    GetSessionInfoActionResult,
)
from ai.backend.manager.services.session.actions.interrupt_session import (
    InterruptSessionAction,
    InterruptSessionActionResult,
)
from ai.backend.manager.services.session.actions.list_files import (
    ListFilesAction,
    ListFilesActionResult,
)
from ai.backend.manager.services.session.actions.match_sessions import (
    MatchSessionsAction,
    MatchSessionsActionResult,
)
from ai.backend.manager.services.session.actions.modify_session import (
    ModifySessionAction,
    ModifySessionActionResult,
)
from ai.backend.manager.services.session.actions.rename_session import (
    RenameSessionAction,
    RenameSessionActionResult,
)
from ai.backend.manager.services.session.actions.restart_session import (
    RestartSessionAction,
    RestartSessionActionResult,
)
from ai.backend.manager.services.session.actions.shutdown_service import (
    ShutdownServiceAction,
    ShutdownServiceActionResult,
)
from ai.backend.manager.services.session.actions.start_service import (
    StartServiceAction,
    StartServiceActionResult,
)
from ai.backend.manager.services.session.actions.upload_files import (
    UploadFilesAction,
    UploadFilesActionResult,
)
from ai.backend.manager.types import UserScope
from ai.backend.manager.utils import query_userinfo

from ...data.image.types import ImageStatus

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore


@dataclass
class SessionServiceArgs:
    db: ExtendedAsyncSAEngine
    agent_registry: AgentRegistry
    background_task_manager: BackgroundTaskManager
    error_monitor: ErrorPluginContext
    idle_checker_host: IdleCheckerHost


class SessionService:
    _db: ExtendedAsyncSAEngine
    _agent_registry: AgentRegistry
    _background_task_manager: BackgroundTaskManager
    _error_monitor: ErrorPluginContext
    _idle_checker_host: IdleCheckerHost
    _database_ptask_group: aiotools.PersistentTaskGroup
    _rpc_ptask_group: aiotools.PersistentTaskGroup

    def __init__(
        self,
        args: SessionServiceArgs,
    ) -> None:
        self._db = args.db
        self._agent_registry = args.agent_registry
        self._background_task_manager = args.background_task_manager
        self._error_monitor = args.error_monitor
        self._idle_checker_host = args.idle_checker_host
        self._database_ptask_group = aiotools.PersistentTaskGroup()
        self._rpc_ptask_group = aiotools.PersistentTaskGroup()
        self._webhook_ptask_group = aiotools.PersistentTaskGroup()

    async def commit_session(self, action: CommitSessionAction) -> CommitSessionActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        filename = action.filename

        myself = asyncio.current_task()
        assert myself is not None

        try:
            async with self._db.begin_readonly_session() as db_sess:
                session = await SessionRow.get_session(
                    db_sess,
                    session_name,
                    owner_access_key,
                    kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
                )

            resp: Mapping[str, Any] = await asyncio.shield(
                self._rpc_ptask_group.create_task(
                    self._agent_registry.commit_session_to_file(session, filename),
                ),
            )
        except BackendError:
            log.exception("COMMIT_SESSION: exception")
            raise

        return CommitSessionActionResult(
            session_row=session,
            commit_result=resp,
        )

    async def complete(self, action: CompleteAction) -> CompleteActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        code = action.code
        options = action.options or {}

        resp = {
            "result": {
                "status": "finished",
                "completions": [],
            },
        }
        async with self._db.begin_readonly_session() as db_sess:
            session = await SessionRow.get_session(
                db_sess,
                session_name,
                owner_access_key,
                kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
            )
        try:
            await self._agent_registry.increment_session_usage(session)
            resp["result"] = cast(
                Dict[str, Any],
                await self._agent_registry.get_completions(session, code, opts=options),
            )
        except AssertionError:
            raise InvalidAPIParameters
        except BackendError:
            log.exception("COMPLETE: exception")
            raise
        return CompleteActionResult(
            session_row=session,
            result=resp,
        )

    async def convert_session_to_image(
        self, action: ConvertSessionToImageAction
    ) -> ConvertSessionToImageActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        image_name = action.image_name
        image_visibility = action.image_visibility
        image_owner_id = action.image_owner_id

        myself = asyncio.current_task()
        assert myself is not None

        if image_visibility != CustomizedImageVisibilityScope.USER:
            raise InvalidAPIParameters(f"Unsupported visibility scope {image_visibility}")

        async with self._db.begin_readonly_session() as db_sess:
            session = await SessionRow.get_session(
                db_sess,
                session_name,
                owner_access_key,
                kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
                eager_loading_op=[selectinload(SessionRow.group)],
            )

        project: GroupRow = session.group
        if not project.container_registry:
            raise InvalidAPIParameters(
                "Project not ready to convert session image (registry configuration not populated)"
            )

        registry_hostname = project.container_registry["registry"]
        registry_project = project.container_registry["project"]

        async with self._db.begin_readonly_session() as db_session:
            query = (
                sa.select(ContainerRegistryRow)
                .where(
                    (ContainerRegistryRow.registry_name == registry_hostname)
                    & (ContainerRegistryRow.project == registry_project)
                )
                .options(
                    load_only(
                        ContainerRegistryRow.url,
                        ContainerRegistryRow.username,
                        ContainerRegistryRow.password,
                        ContainerRegistryRow.project,
                    )
                )
            )

            registry_conf = cast(ContainerRegistryRow | None, await db_session.scalar(query))

            if not registry_conf:
                raise InvalidAPIParameters(
                    f"Project {registry_project} not found in registry {registry_hostname}."
                )

        async with self._db.begin_readonly_session() as db_sess:
            image_row = await ImageRow.resolve(
                db_sess,
                [ImageIdentifier(session.main_kernel.image, session.main_kernel.architecture)],
            )

        base_image_ref = image_row.image_ref

        async def _commit_and_upload(reporter: ProgressReporter) -> None:
            reporter.total_progress = 3
            await reporter.update(message="Commit started")
            try:
                # remove any existing customized related tag from base canonical
                filtered_tag_set = [
                    x for x in base_image_ref.tag.split("-") if not x.startswith("customized_")
                ]

                if base_image_ref.name == "":
                    new_name = base_image_ref.project
                else:
                    new_name = base_image_ref.name

                new_canonical = f"{registry_hostname}/{registry_project}/{new_name}:{'-'.join(filtered_tag_set)}"

                async with self._db.begin_readonly_session() as sess:
                    # check if user has passed its limit of customized image count
                    query = (
                        sa.select([sa.func.count()])
                        .select_from(ImageRow)
                        .where(
                            (
                                ImageRow.labels["ai.backend.customized-image.owner"].as_string()
                                == f"{image_visibility.value}:{image_owner_id}"
                            )
                        )
                        .where(ImageRow.status == ImageStatus.ALIVE)
                    )
                    existing_image_count = await sess.scalar(query)

                    customized_image_count_limit = action.max_customized_image_count
                    if customized_image_count_limit <= existing_image_count:
                        raise QuotaExceeded(
                            extra_msg="You have reached your customized image count quota",
                            extra_data={
                                "limit": customized_image_count_limit,
                                "current": existing_image_count,
                            },
                        )

                    # check if image with same name exists and reuse ID it if is
                    query = sa.select(ImageRow).where(
                        sa.and_(
                            ImageRow.name.like(f"{new_canonical}%"),
                            ImageRow.labels["ai.backend.customized-image.owner"].as_string()
                            == f"{image_visibility.value}:{image_owner_id}",
                            ImageRow.labels["ai.backend.customized-image.name"].as_string()
                            == image_name,
                            ImageRow.status == ImageStatus.ALIVE,
                        )
                    )
                    existing_row = await sess.scalar(query)

                    customized_image_id: str
                    kern_features: list[str]
                    if existing_row:
                        kern_features = existing_row.labels.get(
                            LabelName.FEATURES.value, DEFAULT_KERNEL_FEATURE
                        ).split()
                        customized_image_id = existing_row.labels[LabelName.CUSTOMIZED_ID.value]
                        log.debug("reusing existing customized image ID {}", customized_image_id)
                    else:
                        kern_features = [DEFAULT_KERNEL_FEATURE]
                        customized_image_id = str(uuid.uuid4())
                    # Remove PRIVATE label for customized images
                    kern_features = [
                        feat for feat in kern_features if feat != KernelFeatures.PRIVATE.value
                    ]

                new_canonical += f"-customized_{customized_image_id.replace('-', '')}"
                new_image_ref = ImageRef.from_image_str(
                    new_canonical,
                    None,
                    registry_hostname,
                    architecture=base_image_ref.architecture,
                    is_local=base_image_ref.is_local,
                )

                image_labels = {
                    LabelName.CUSTOMIZED_OWNER.value: f"{image_visibility.value}:{image_owner_id}",
                    LabelName.CUSTOMIZED_NAME.value: image_name,
                    LabelName.CUSTOMIZED_ID.value: customized_image_id,
                    LabelName.FEATURES.value: " ".join(kern_features),
                }
                match image_visibility:
                    case CustomizedImageVisibilityScope.USER:
                        image_labels[LabelName.CUSTOMIZED_USER_EMAIL.value] = action.user_email

                # commit image with new tag set
                resp = await self._agent_registry.commit_session(
                    session,
                    new_image_ref,
                    extra_labels=image_labels,
                )
                async for event, _ in self._background_task_manager.poll_bgtask_event(
                    uuid.UUID(resp["bgtask_id"])
                ):
                    match event:
                        case BgtaskDoneEvent():
                            await reporter.update(increment=1, message="Committed image")
                            break
                        case BgtaskFailedEvent():
                            raise BackendError(extra_msg=event.message)
                        case BgtaskCancelledEvent():
                            raise BackendError(extra_msg="Operation cancelled")

                if not new_image_ref.is_local:
                    # push image to registry from local agent
                    image_registry = ImageRegistry(
                        name=registry_hostname,
                        url=str(registry_conf.url),
                        username=registry_conf.username,
                        password=registry_conf.password,
                    )
                    resp = await self._agent_registry.push_image(
                        session.main_kernel.agent,
                        new_image_ref,
                        image_registry,
                    )
                    async for event, _ in self._background_task_manager.poll_bgtask_event(
                        uuid.UUID(resp["bgtask_id"])
                    ):
                        match event:
                            case BgtaskDoneEvent():
                                break
                            case BgtaskFailedEvent():
                                raise BackendError(extra_msg=event.message)
                            case BgtaskCancelledEvent():
                                raise BackendError(extra_msg="Operation cancelled")

                await reporter.update(increment=1, message="Pushed image to registry")
                # rescan updated image only
                await rescan_images(
                    self._db,
                    new_image_ref.canonical,
                    registry_project,
                    reporter=reporter,
                )
                await reporter.update(increment=1, message="Completed")
            except BackendError:
                log.exception("CONVERT_SESSION_TO_IMAGE: exception")
                raise

        task_id = await self._background_task_manager.start(_commit_and_upload)
        return ConvertSessionToImageActionResult(task_id=task_id, session_row=session)

    async def create_cluster(self, action: CreateClusterAction) -> CreateClusterActionResult:
        template_id = action.template_id
        user_id = action.user_id
        user_role = action.user_role
        sudo_session_enabled = action.sudo_session_enabled
        keypair_resource_policy = action.keypair_resource_policy
        requester_access_key = action.requester_access_key
        owner_access_key = action.owner_access_key
        domain_name = action.domain_name
        group_name = action.group_name
        scaling_group_name = action.scaling_group_name
        session_name = action.session_name
        session_type = action.session_type
        enqueue_only = action.enqueue_only
        max_wait_seconds = action.max_wait_seconds
        tag = action.tag

        async with self._db.begin_readonly() as conn:
            query = (
                sa.select([session_templates.c.template])
                .select_from(session_templates)
                .where(
                    (session_templates.c.id == template_id) & session_templates.c.is_active,
                )
            )
            template = await conn.scalar(query)
            log.debug("task template: {}", template)
            if not template:
                raise TaskTemplateNotFound

            try:
                _owner_uuid, group_id, resource_policy = await query_userinfo(
                    conn,
                    user_id,
                    requester_access_key,
                    user_role,
                    domain_name,
                    keypair_resource_policy,
                    domain_name,
                    group_name,
                    query_on_behalf_of=(
                        None if owner_access_key is undefined else owner_access_key
                    ),
                )

            except ValueError as e:
                raise InvalidAPIParameters(str(e))

        try:
            resp = await self._agent_registry.create_cluster(
                template,
                session_name,
                UserScope(
                    domain_name=domain_name,
                    group_id=group_id,
                    user_uuid=user_id,
                    user_role=user_role,
                ),
                owner_access_key,
                resource_policy,
                scaling_group_name,
                session_type,
                tag,
                enqueue_only=enqueue_only,
                max_wait_seconds=max_wait_seconds,
                sudo_session_enabled=sudo_session_enabled,
            )
            return CreateClusterActionResult(result=resp, session_id=resp["kernelId"])
        except TooManySessionsMatched:
            raise SessionAlreadyExists
        except BackendError:
            log.exception("GET_OR_CREATE: exception")
            raise
        except UnknownImageReference:
            raise UnknownImageReferenceError("Unknown image reference!")
        except Exception:
            await self._error_monitor.capture_exception()
            log.exception("GET_OR_CREATE: unexpected error!")
            raise InternalServerError

    async def create_from_params(
        self, action: CreateFromParamsAction
    ) -> CreateFromParamsActionResult:
        user_id = action.user_id
        user_role = action.user_role
        sudo_session_enabled = action.sudo_session_enabled
        keypair_resource_policy = action.keypair_resource_policy
        requester_access_key = action.requester_access_key

        owner_access_key = action.params.owner_access_key
        domain_name = action.params.domain_name
        group_name = action.params.group_name
        config = action.params.config
        cluster_size = action.params.cluster_size
        cluster_mode = action.params.cluster_mode
        session_name = action.params.session_name
        session_type = action.params.session_type
        enqueue_only = action.params.enqueue_only
        max_wait_seconds = action.params.max_wait_seconds
        tag = action.params.tag
        image = action.params.image
        architecture = action.params.architecture
        priority = action.params.priority
        bootstrap_script = action.params.bootstrap_script
        dependencies = action.params.dependencies
        startup_command = action.params.startup_command
        starts_at = action.params.starts_at
        batch_timeout = action.params.batch_timeout
        callback_url = action.params.callback_url
        reuse_if_exists = action.params.reuse_if_exists

        async with self._db.begin_readonly() as conn:
            owner_uuid, group_id, resource_policy = await query_userinfo(
                conn,
                user_id,
                requester_access_key,
                user_role,
                domain_name,
                keypair_resource_policy,
                domain_name,
                group_name,
                query_on_behalf_of=(None if owner_access_key is undefined else owner_access_key),
            )

        try:
            async with self._db.begin_session() as session:
                image_row = await ImageRow.resolve(
                    session,
                    [
                        ImageIdentifier(
                            image,
                            architecture,
                        ),
                        ImageAlias(image),
                    ],
                )

            resp = await self._agent_registry.create_session(
                session_name,
                image_row.image_ref,
                UserScope(
                    domain_name=domain_name,
                    group_id=group_id,
                    user_uuid=user_id,
                    user_role=user_role,
                ),
                owner_access_key,
                resource_policy,
                session_type,
                config,
                cluster_mode,
                cluster_size,
                reuse=reuse_if_exists,
                priority=priority,
                enqueue_only=enqueue_only,
                max_wait_seconds=max_wait_seconds,
                bootstrap_script=bootstrap_script,
                dependencies=dependencies,
                startup_command=startup_command,
                starts_at_timestamp=starts_at,
                batch_timeout=batch_timeout,
                tag=tag,
                callback_url=callback_url,
                sudo_session_enabled=sudo_session_enabled,
            )
            return CreateFromParamsActionResult(session_id=resp["sessionId"], result=resp)
        except UnknownImageReference:
            raise UnknownImageReferenceError(f"Unknown image reference: {image}")
        except BackendError:
            log.exception("GET_OR_CREATE: exception")
            raise
        except Exception:
            await self._error_monitor.capture_exception(context={"user": owner_uuid})
            log.exception("GET_OR_CREATE: unexpected error!")
            raise InternalServerError

    async def create_from_template(
        self, action: CreateFromTemplateAction
    ) -> CreateFromTemplateActionResult:
        template_id = action.params.template_id
        user_id = action.user_id
        user_role = action.user_role
        sudo_session_enabled = action.sudo_session_enabled
        keypair_resource_policy = action.keypair_resource_policy
        requester_access_key = action.requester_access_key

        async with self._db.begin_readonly() as conn:
            query = (
                sa.select([session_templates])
                .select_from(session_templates)
                .where(
                    (session_templates.c.id == template_id) & session_templates.c.is_active,
                )
            )
            result = await conn.execute(query)
            template_info = result.fetchone()
            template = template_info["template"]
            if not template:
                raise TaskTemplateNotFound

            group_name = None
            if template_info["domain_name"] and template_info["group_id"]:
                query = (
                    sa.select([groups.c.name])
                    .select_from(groups)
                    .where(
                        (groups.c.domain_name == template_info["domain_name"])
                        & (groups.c.id == template_info["group_id"]),
                    )
                )
                group_name = await conn.scalar(query)

        if isinstance(template, str):
            template = load_json(template)

        log.debug("Template: {0}", template)

        param_from_template = {
            "image": template["spec"]["kernel"]["image"],
            "architecture": template["spec"]["kernel"]["architecture"],
        }

        if "domain_name" in template_info:
            param_from_template["domain"] = template_info["domain_name"]
        if group_name:
            param_from_template["group"] = group_name
        if template["spec"]["session_type"] == "interactive":
            param_from_template["session_type"] = SessionTypes.INTERACTIVE
        elif template["spec"]["session_type"] == "batch":
            param_from_template["session_type"] = SessionTypes.BATCH
        elif template["spec"]["session_type"] == "inference":
            param_from_template["session_type"] = SessionTypes.INFERENCE

        if tag := template["metadata"].get("tag"):
            param_from_template["tag"] = tag
        if runtime_opt := template["spec"]["kernel"]["run"]:
            if bootstrap := runtime_opt["bootstrap"]:
                param_from_template["bootstrap_script"] = bootstrap
            if startup := runtime_opt["startup_command"]:
                param_from_template["startup_command"] = startup

        config_from_template: MutableMapping[Any, Any] = {}
        if scaling_group := template["spec"].get("scaling_group"):
            config_from_template["scaling_group"] = scaling_group
        if mounts := template["spec"].get("mounts"):
            config_from_template["mounts"] = list(mounts.keys())
            config_from_template["mount_map"] = {
                key: value for (key, value) in mounts.items() if len(value) > 0
            }
        if environ := template["spec"]["kernel"].get("environ"):
            config_from_template["environ"] = environ
        if resources := template["spec"].get("resources"):
            config_from_template["resources"] = resources
        if "agent_list" in template["spec"]:
            config_from_template["agent_list"] = template["spec"]["agent_list"]

        override_config = drop_undefined(dict(action.params.config))
        override_params = drop_undefined(dict(asdict(action.params)))

        log.debug("Default config: {0}", config_from_template)
        log.debug("Default params: {0}", param_from_template)

        log.debug("Override config: {0}", override_config)
        log.debug("Override params: {0}", override_params)

        if override_config:
            config_from_template.update(override_config)
        if override_params:
            param_from_template.update(override_params)

        try:
            params = overwritten_param_check.check(param_from_template)
        except RuntimeError as e1:
            log.exception(e1)
        except t.DataError as e2:
            log.debug("Error: {0}", str(e2))
            raise InvalidAPIParameters("Error while validating template")
        params["config"] = config_from_template

        log.debug("Updated param: {0}", params)

        if git := template["spec"]["kernel"]["git"]:
            if _dest := git.get("dest_dir"):
                target = _dest
            else:
                target = git["repository"].split("/")[-1]

            cmd_builder = "git clone "
            if credential := git.get("credential"):
                proto, url = git["repository"].split("://")
                cmd_builder += f"{proto}://{credential['username']}:{credential['password']}@{url}"
            else:
                cmd_builder += git["repository"]
            if branch := git.get("branch"):
                cmd_builder += f" -b {branch}"
            cmd_builder += f" {target}\n"

            if commit := git.get("commit"):
                cmd_builder = "CWD=$(pwd)\n" + cmd_builder
                cmd_builder += f"cd {target}\n"
                cmd_builder += f"git checkout {commit}\n"
                cmd_builder += "cd $CWD\n"

            bootstrap = base64.b64decode(params.get("bootstrap_script") or b"").decode()
            bootstrap += "\n"
            bootstrap += cmd_builder
            params["bootstrap_script"] = base64.b64encode(bootstrap.encode()).decode()

        owner_access_key = params["owner_access_key"]
        config = params["config"]
        cluster_size = params["cluster_size"]
        cluster_mode = params["cluster_mode"]
        session_name = params["session_name"]
        session_type = params["session_type"]
        enqueue_only = params["enqueue_only"]
        max_wait_seconds = params["max_wait_seconds"]
        tag = params["tag"]
        image = params["image"]
        architecture = params["architecture"]
        priority = params["priority"]
        bootstrap_script = params["bootstrap_script"]
        dependencies = params["dependencies"]
        startup_command = params["startup_command"]
        starts_at = params["starts_at"]
        batch_timeout = params["batch_timeout"]
        callback_url = params["callback_url"]
        reuse_if_exists = params["reuse_if_exists"]
        domain_name = params["domain_name"]

        async with self._db.begin_readonly() as conn:
            owner_uuid, group_id, resource_policy = await query_userinfo(
                conn,
                user_id,
                requester_access_key,
                user_role,
                domain_name,
                keypair_resource_policy,
                domain_name,
                params["group_name"],
                query_on_behalf_of=(None if owner_access_key is undefined else owner_access_key),
            )

        try:
            async with self._db.begin_session() as session:
                image_row = await ImageRow.resolve(
                    session,
                    [
                        ImageIdentifier(
                            image,
                            architecture,
                        ),
                        ImageAlias(image),
                    ],
                )

            resp = await self._agent_registry.create_session(
                session_name,
                image_row.image_ref,
                UserScope(
                    domain_name=domain_name,
                    group_id=group_id,
                    user_uuid=user_id,
                    user_role=user_role,
                ),
                owner_access_key,
                resource_policy,
                session_type,
                config,
                cluster_mode,
                cluster_size,
                reuse=reuse_if_exists,
                priority=priority,
                enqueue_only=enqueue_only,
                max_wait_seconds=max_wait_seconds,
                bootstrap_script=bootstrap_script,
                dependencies=dependencies,
                startup_command=startup_command,
                starts_at_timestamp=starts_at,
                batch_timeout=batch_timeout,
                tag=tag,
                callback_url=callback_url,
                sudo_session_enabled=sudo_session_enabled,
            )
            return CreateFromTemplateActionResult(session_id=resp["sessionId"], result=resp)
        except UnknownImageReference:
            raise UnknownImageReferenceError(f"Unknown image reference: {image}")
        except BackendError:
            log.exception("GET_OR_CREATE: exception")
            raise
        except Exception:
            await self._error_monitor.capture_exception(context={"user": owner_uuid})
            log.exception("GET_OR_CREATE: unexpected error!")
            raise InternalServerError

    async def destroy_session(self, action: DestroySessionAction) -> DestroySessionActionResult:
        user_role = action.user_role
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        forced = action.forced
        recursive = action.recursive

        if recursive:
            async with self._db.begin_readonly_session() as db_sess:
                dependent_session_ids = await find_dependent_sessions(
                    session_name,
                    db_sess,
                    owner_access_key,
                    allow_stale=True,
                )

                target_session_references: list[str | uuid.UUID] = [
                    *dependent_session_ids,
                    session_name,
                ]
                sessions: Iterable[SessionRow | BaseException] = await asyncio.gather(
                    *[
                        SessionRow.get_session(
                            db_sess,
                            name_or_id,
                            owner_access_key,
                            kernel_loading_strategy=KernelLoadingStrategy.ALL_KERNELS,
                        )
                        for name_or_id in target_session_references
                    ],
                    return_exceptions=True,
                )

            last_stats = await asyncio.gather(
                *[
                    self._agent_registry.destroy_session(sess, forced=forced, user_role=user_role)
                    for sess in sessions
                    if isinstance(sess, SessionRow)
                ],
                return_exceptions=True,
            )

            # Consider not found sessions already terminated.
            # Consider GenericForbidden error occurs with scheduled/preparing/terminating/error status session, and leave them not to be quitted.
            last_stats = [
                *filter(lambda x: not isinstance(x, SessionNotFound | GenericForbidden), last_stats)
            ]

            return DestroySessionActionResult(result=last_stats, destroyed_sessions=sessions)
        else:
            async with self._db.begin_readonly_session() as db_sess:
                session = await SessionRow.get_session(
                    db_sess,
                    session_name,
                    owner_access_key,
                    kernel_loading_strategy=KernelLoadingStrategy.ALL_KERNELS,
                )
            last_stat = await self._agent_registry.destroy_session(
                session,
                forced=forced,
                user_role=user_role,
            )
            resp = {
                "stats": last_stat,
            }

            return DestroySessionActionResult(result=resp, destroyed_sessions=[session])

    async def download_file(self, action: DownloadFileAction) -> DownloadFileActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        user_id = action.user_id
        file = action.file

        try:
            async with self._db.begin_readonly_session() as db_sess:
                session = await SessionRow.get_session(
                    db_sess,
                    session_name,
                    owner_access_key,
                    kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
                )
            await self._agent_registry.increment_session_usage(session)
            result = await self._agent_registry.download_single(session, owner_access_key, file)
        except asyncio.CancelledError:
            raise
        except BackendError:
            log.exception("DOWNLOAD_SINGLE: exception")
            raise
        except (ValueError, FileNotFoundError):
            raise InvalidAPIParameters("The file is not found.")
        except Exception:
            await self._error_monitor.capture_exception(context={"user": user_id})
            log.exception("DOWNLOAD_SINGLE: unexpected error!")
            raise InternalServerError

        return DownloadFileActionResult(result=result, session_row=session)

    async def download_files(self, action: DownloadFilesAction) -> DownloadFilesActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        user_id = action.user_id
        files = action.files

        async with self._db.begin_readonly_session() as db_sess:
            session = await SessionRow.get_session(
                db_sess,
                session_name,
                owner_access_key,
                kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
            )
        try:
            assert len(files) <= 5, "Too many files"
            await self._agent_registry.increment_session_usage(session)
            # TODO: Read all download file contents. Need to fix by using chuncking, etc.
            results = await asyncio.gather(
                *map(
                    functools.partial(self._agent_registry.download_file, session),
                    files,
                ),
            )
            log.debug("file(s) inside container retrieved")
        except asyncio.CancelledError:
            raise
        except BackendError:
            log.exception("DOWNLOAD_FILE: exception")
            raise
        except (ValueError, FileNotFoundError):
            raise InvalidAPIParameters("The file is not found.")
        except Exception:
            await self._error_monitor.capture_exception(context={"user": user_id})
            log.exception("DOWNLOAD_FILE: unexpected error!")
            raise InternalServerError

        with aiohttp.MultipartWriter("mixed") as mpwriter:
            headers = multidict.MultiDict({"Content-Encoding": "identity"})
            for tarbytes in results:
                mpwriter.append(tarbytes, headers)

            return DownloadFilesActionResult(
                session_row=session,
                result=mpwriter,  # type: ignore
            )

    async def execute_session(self, action: ExecuteSessionAction) -> ExecuteSessionActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        api_version = action.api_version

        resp = {}
        async with self._db.begin_readonly_session() as db_sess:
            session = await SessionRow.get_session(
                db_sess,
                session_name,
                owner_access_key,
                kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
            )
        try:
            await self._agent_registry.increment_session_usage(session)

            run_id: Optional[str]
            mode: Optional[str]

            if api_version[0] == 1:
                run_id = action.params.run_id or secrets.token_hex(8)
                mode = "query"
                code = action.params.code
                opts = None
            elif api_version[0] >= 2:
                run_id = action.params.run_id
                mode = action.params.mode

                if mode is None:
                    # TODO: Create new exception
                    raise RuntimeError("runId or mode is missing!")

                assert mode in {
                    "query",
                    "batch",
                    "complete",
                    "continue",
                    "input",
                }, "mode has an invalid value."
                if mode in {"continue", "input"}:
                    assert run_id is not None, "continuation requires explicit run ID"
                code = action.params.code
                opts = action.params.options
            else:
                raise RuntimeError("should not reach here")
            # handle cases when some params are deliberately set to None
            if code is None:
                code = ""  # noqa
            if opts is None:
                opts = {}  # noqa
            if mode == "complete":
                # For legacy
                resp["result"] = await self._agent_registry.get_completions(session, code, opts)
            else:
                run_id = cast(str, run_id)
                raw_result = await self._agent_registry.execute(
                    session,
                    api_version,
                    run_id,
                    mode,
                    code,
                    opts,
                    flush_timeout=2.0,
                )
                if raw_result is None:
                    # the kernel may have terminated from its side,
                    # or there was interruption of agents.
                    resp["result"] = {
                        "status": "finished",
                        "runId": run_id,
                        "exitCode": 130,
                        "options": {},
                        "files": [],
                        "console": [],
                    }
                    return ExecuteSessionActionResult(result=resp, session_row=session)
                # Keep internal/public API compatilibty
                result = {
                    "status": raw_result["status"],
                    "runId": raw_result["runId"],
                    "exitCode": raw_result.get("exitCode"),
                    "options": raw_result.get("options"),
                    "files": raw_result.get("files"),
                }
                if api_version[0] == 1:
                    result["stdout"] = raw_result.get("stdout")
                    result["stderr"] = raw_result.get("stderr")
                    result["media"] = raw_result.get("media")
                    result["html"] = raw_result.get("html")
                else:
                    result["console"] = raw_result.get("console")
                resp["result"] = result
        except AssertionError as e:
            log.warning("EXECUTE: invalid/missing parameters: {0!r}", e)
            raise InvalidAPIParameters(extra_msg=e.args[0])
        except BackendError:
            log.exception("EXECUTE: exception")
            raise

        return ExecuteSessionActionResult(result=resp, session_row=session)

    async def get_abusing_report(
        self, action: GetAbusingReportAction
    ) -> GetAbusingReportActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        try:
            async with self._db.begin_readonly_session() as db_sess:
                session = await SessionRow.get_session(
                    db_sess,
                    session_name,
                    owner_access_key,
                    kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
                )
            kernel = session.main_kernel
            report = await self._agent_registry.get_abusing_report(kernel.id)
        except BackendError:
            log.exception("GET_ABUSING_REPORT: exception")
            raise
        return GetAbusingReportActionResult(result=report, session_row=session)

    async def get_commit_status(self, action: GetCommitStatusAction) -> GetCommitStatusActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key

        try:
            async with self._db.begin_readonly_session() as db_sess:
                session = await SessionRow.get_session(
                    db_sess,
                    session_name,
                    owner_access_key,
                    kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
                )
            statuses = await self._agent_registry.get_commit_status([session.main_kernel.id])
        except BackendError:
            log.exception("GET_COMMIT_STATUS: exception")
            raise
        resp = {"status": statuses[session.main_kernel.id], "kernel": str(session.main_kernel.id)}
        return GetCommitStatusActionResult(result=resp, session_row=session)

    async def get_container_logs(
        self, action: GetContainerLogsAction
    ) -> GetContainerLogsActionResult:
        resp = {"result": {"logs": ""}}
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        kernel_id = action.kernel_id

        async with self._db.begin_readonly_session() as db_sess:
            compute_session = await SessionRow.get_session(
                db_sess,
                session_name,
                owner_access_key,
                allow_stale=True,
                kernel_loading_strategy=(
                    KernelLoadingStrategy.MAIN_KERNEL_ONLY
                    if kernel_id is None
                    else KernelLoadingStrategy.ALL_KERNELS
                ),
            )

            if compute_session.status in DEAD_SESSION_STATUSES:
                if kernel_id is None:
                    # Get logs from the main kernel
                    kernel_id = compute_session.main_kernel.id
                    kernel_log = compute_session.main_kernel.container_log
                else:
                    # Get logs from the specific kernel
                    kernel_row = compute_session.get_kernel_by_id(kernel_id)
                    kernel_log = kernel_row.container_log
                if kernel_log is not None:
                    # Get logs from database record
                    log.debug("returning log from database record")
                    resp["result"]["logs"] = kernel_log.decode("utf-8")
                    return GetContainerLogsActionResult(result=resp, session_row=compute_session)

        registry = self._agent_registry
        await registry.increment_session_usage(compute_session)
        resp["result"]["logs"] = await registry.get_logs_from_agent(
            session=compute_session, kernel_id=kernel_id
        )
        log.debug("returning log from agent")

        return GetContainerLogsActionResult(result=resp, session_row=compute_session)

    async def get_dependency_graph(
        self, action: GetDependencyGraphAction
    ) -> GetDependencyGraphActionResult:
        root_session_name = action.root_session_name
        owner_access_key = action.owner_access_key

        async with self._db.begin_readonly_session() as db_session:
            # TODO: Move `find_dependency_sessions` impl to Service layer
            dependency_graph = await find_dependency_sessions(
                root_session_name, db_session, owner_access_key
            )
            session_id = dependency_graph["session_id"]
            stmt = sa.select(SessionRow).where(SessionRow.id == session_id)
            session = await db_session.scalar(stmt)

        return GetDependencyGraphActionResult(result=dependency_graph, session_row=session)

    async def get_direct_access_info(
        self, action: GetDirectAccessInfoAction
    ) -> GetDirectAccessInfoActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key

        async with self._db.begin_session() as db_sess:
            sess = await SessionRow.get_session(
                db_sess,
                session_name,
                owner_access_key,
                kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
            )
        resp = {}
        sess_type = cast(SessionTypes, sess.session_type)
        if sess_type in PRIVATE_SESSION_TYPES:
            public_host = sess.main_kernel.agent_row.public_host
            found_ports: dict[str, list[str]] = {}
            service_ports = cast(Optional[list[dict[str, Any]]], sess.main_kernel.service_ports)
            if service_ports is None:
                raise KernelNotReady(
                    f"Kernel of the session has no service ports yet (kernel: {sess.main_kernel.id}, kernel status: {sess.main_kernel.status.name})"
                )
            for sport in service_ports:
                if sport["name"] == "sshd":
                    found_ports["sshd"] = sport["host_ports"]
                elif sport["name"] == "sftpd":
                    found_ports["sftpd"] = sport["host_ports"]
            resp = {
                "kernel_role": sess_type.name,  # legacy
                "session_type": sess_type.name,
                "public_host": public_host,
                "sshd_ports": found_ports.get("sftpd") or found_ports["sshd"],
            }

        return GetDirectAccessInfoActionResult(result=resp, session_row=sess)

    async def get_session_info(self, action: GetSessionInfoAction) -> GetSessionInfoActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key

        resp = {}
        async with self._db.begin_session() as db_sess:
            sess = await SessionRow.get_session(
                db_sess,
                session_name,
                owner_access_key,
                kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
            )
        await self._agent_registry.increment_session_usage(sess)
        resp["domainName"] = sess.domain_name
        resp["groupId"] = str(sess.group_id)
        resp["userId"] = str(sess.user_uuid)
        resp["lang"] = sess.main_kernel.image  # legacy
        resp["image"] = sess.main_kernel.image
        resp["architecture"] = sess.main_kernel.architecture
        resp["registry"] = sess.main_kernel.registry
        resp["tag"] = sess.tag

        # Resource occupation
        resp["containerId"] = str(sess.main_kernel.container_id)
        resp["occupiedSlots"] = str(sess.main_kernel.occupied_slots)  # legacy
        resp["occupyingSlots"] = str(sess.occupying_slots)
        resp["requestedSlots"] = str(sess.requested_slots)
        resp["occupiedShares"] = str(
            sess.main_kernel.occupied_shares
        )  # legacy, only caculate main kernel's occupying resource
        resp["environ"] = str(sess.environ)
        resp["resourceOpts"] = str(sess.resource_opts)

        # Lifecycle
        resp["status"] = sess.status.name  # "e.g. 'SessionStatus.RUNNING' -> 'RUNNING' "
        resp["statusInfo"] = str(sess.status_info)
        resp["statusData"] = sess.status_data
        age = datetime.now(tzutc()) - sess.created_at
        resp["age"] = int(age.total_seconds() * 1000)  # age in milliseconds
        resp["creationTime"] = str(sess.created_at)
        resp["terminationTime"] = str(sess.terminated_at) if sess.terminated_at else None

        resp["numQueriesExecuted"] = sess.num_queries
        resp["lastStat"] = sess.last_stat
        resp["idleChecks"] = await self._idle_checker_host.get_idle_check_report(sess.id)

        # Resource limits collected from agent heartbeats were erased, as they were deprecated
        # TODO: factor out policy/image info as a common repository

        return GetSessionInfoActionResult(result=resp, session_row=sess)

    async def interrupt(self, action: InterruptSessionAction) -> InterruptSessionActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key

        async with self._db.begin_readonly_session() as db_sess:
            session = await SessionRow.get_session(
                db_sess,
                session_name,
                owner_access_key,
                kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
            )
        await self._agent_registry.increment_session_usage(session)
        await self._agent_registry.interrupt_session(session)

        return InterruptSessionActionResult(result=None, session_row=session)

    async def list_files(self, action: ListFilesAction) -> ListFilesActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        user_id = action.user_id
        path = action.path

        async with self._db.begin_readonly_session() as db_sess:
            session = await SessionRow.get_session(
                db_sess,
                session_name,
                owner_access_key,
                kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
            )

        resp: MutableMapping[str, Any] = {}
        try:
            await self._agent_registry.increment_session_usage(session)
            result = await self._agent_registry.list_files(session, path)
            resp.update(result)
            log.debug("container file list for {0} retrieved", path)
        except asyncio.CancelledError:
            raise
        except BackendError:
            log.exception("LIST_FILES: exception")
            raise
        except Exception:
            await self._error_monitor.capture_exception(context={"user": user_id})
            log.exception("LIST_FILES: unexpected error!")
            raise InternalServerError

        return ListFilesActionResult(result=result, session_row=session)

    async def match_sessions(self, action: MatchSessionsAction) -> MatchSessionsActionResult:
        id_or_name_prefix = action.id_or_name_prefix
        owner_access_key = action.owner_access_key

        matches: list[dict[str, Any]] = []
        async with self._db.begin_readonly_session() as db_sess:
            sessions = await SessionRow.match_sessions(
                db_sess,
                id_or_name_prefix,
                owner_access_key,
            )
        if sessions:
            matches.extend(
                {
                    "id": str(item.id),
                    "name": item.name,
                    "status": item.status.name,
                }
                for item in sessions
            )
        return MatchSessionsActionResult(result=matches)

    async def rename_session(self, action: RenameSessionAction) -> RenameSessionActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        new_name = action.new_name

        async with self._db.begin_session() as db_sess:
            compute_session = await SessionRow.get_session(
                db_sess,
                session_name,
                owner_access_key,
                allow_stale=True,
                kernel_loading_strategy=KernelLoadingStrategy.ALL_KERNELS,
            )
            if compute_session.status != SessionStatus.RUNNING:
                raise InvalidAPIParameters("Can't change name of not running session")
            compute_session.name = new_name
            for kernel in compute_session.kernels:
                kernel.session_name = new_name
            await db_sess.commit()

        return RenameSessionActionResult(result=None, session_row=compute_session)

    async def restart_session(self, action: RestartSessionAction) -> RestartSessionActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key

        async with self._db.begin_session() as db_sess:
            session = await SessionRow.get_session(
                db_sess,
                session_name,
                owner_access_key,
                kernel_loading_strategy=KernelLoadingStrategy.ALL_KERNELS,
            )
        await self._agent_registry.increment_session_usage(session)
        await self._agent_registry.restart_session(session)
        return RestartSessionActionResult(result=None, session_row=session)

    async def shutdown_service(self, action: ShutdownServiceAction) -> ShutdownServiceActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        service_name = action.service_name

        async with self._db.begin_readonly_session() as db_sess:
            session = await SessionRow.get_session(
                db_sess,
                session_name,
                owner_access_key,
                kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
            )
        await self._agent_registry.shutdown_service(session, service_name)
        return ShutdownServiceActionResult(result=None, session_row=session)

    async def start_service(self, action: StartServiceAction) -> StartServiceActionResult:
        session_name = action.session_name
        access_key = action.access_key
        service = action.service
        port = action.port

        arguments = action.arguments
        envs = action.envs
        login_session_token = action.login_session_token

        try:
            async with self._db.begin_readonly_session() as db_sess:
                session = await asyncio.shield(
                    self._database_ptask_group.create_task(
                        SessionRow.get_session(
                            db_sess,
                            session_name,
                            access_key,
                            kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
                            eager_loading_op=[
                                selectinload(SessionRow.routing).options(noload("*")),
                            ],
                        ),
                    )
                )
        except (SessionNotFound, TooManySessionsMatched):
            raise

        query = (
            sa.select([scaling_groups.c.wsproxy_addr])
            .select_from(scaling_groups)
            .where((scaling_groups.c.name == session.scaling_group_name))
        )

        async with self._db.begin_readonly() as conn:
            result = await conn.execute(query)
            sgroup = result.first()
        wsproxy_addr = sgroup["wsproxy_addr"]
        if not wsproxy_addr:
            raise ServiceUnavailable("No coordinator configured for this resource group")
        wsproxy_status = await query_wsproxy_status(wsproxy_addr)
        if advertise_addr := wsproxy_status.get("advertise_address"):
            wsproxy_advertise_addr = advertise_addr
        else:
            wsproxy_advertise_addr = wsproxy_addr

        if session.main_kernel.kernel_host is None:
            kernel_host = urlparse(session.main_kernel.agent_addr).hostname
        else:
            kernel_host = session.main_kernel.kernel_host
        for sport in session.main_kernel.service_ports:
            if sport["name"] == service:
                if sport["is_inference"]:
                    raise InvalidAPIParameters(
                        f"{service} is an inference app. Starting inference apps can only be done by"
                        " starting an inference service."
                    )
                if port:
                    # using one of the primary/secondary ports of the app
                    try:
                        hport_idx = sport["container_ports"].index(port)
                    except ValueError:
                        raise InvalidAPIParameters(
                            f"Service {service} does not open the port number {port}."
                        )
                    host_port = sport["host_ports"][hport_idx]
                else:
                    # using the default (primary) port of the app
                    if "host_ports" not in sport:
                        host_port = sport["host_port"]  # legacy kernels
                    else:
                        host_port = sport["host_ports"][0]
                break
        else:
            raise AppNotFound(f"{session_name}:{service}")

        await asyncio.shield(
            self._database_ptask_group.create_task(
                self._agent_registry.increment_session_usage(session),
            )
        )

        opts: MutableMapping[str, Union[None, str, list[str]]] = {}
        if arguments is not None:
            opts["arguments"] = load_json(arguments)
        if envs is not None:
            opts["envs"] = load_json(envs)

        result = await asyncio.shield(
            self._rpc_ptask_group.create_task(
                self._agent_registry.start_service(session, service, opts),
            ),
        )
        if result["status"] == "failed":
            raise InternalServerError(
                "Failed to launch the app service", extra_data=result["error"]
            )

        body = {
            "login_session_token": login_session_token,
            "kernel_host": kernel_host,
            "kernel_port": host_port,
            "session": {
                "id": str(session.id),
                "user_uuid": str(session.user_uuid),
                "group_id": str(session.group_id),
                "access_key": session.access_key,
                "domain_name": session.domain_name,
            },
        }

        async with aiohttp.ClientSession() as req:
            async with req.post(
                f"{wsproxy_addr}/v2/conf",
                json=body,
            ) as resp:
                token_json = await resp.json()

                return StartServiceActionResult(
                    result=None,
                    session_row=session,
                    token=token_json["token"],
                    wsproxy_addr=wsproxy_advertise_addr,
                )

    async def upload_files(self, action: UploadFilesAction) -> UploadFilesActionResult:
        session_name = action.session_name
        owner_access_key = action.owner_access_key
        reader = action.reader

        loop = asyncio.get_event_loop()

        async with self._db.begin_readonly_session() as db_sess:
            session = await SessionRow.get_session(
                db_sess,
                session_name,
                owner_access_key,
                kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
            )

        await self._agent_registry.increment_session_usage(session)
        file_count = 0
        upload_tasks = []
        async for file in aiotools.aiter(reader.next, None):
            if file_count == 20:
                raise InvalidAPIParameters("Too many files")
            file_count += 1
            # This API handles only small files, so let's read it at once.
            chunks = []
            recv_size = 0
            while True:
                chunk = await file.read_chunk(size=1048576)
                if not chunk:
                    break
                chunk_size = len(chunk)
                if recv_size + chunk_size >= 1048576:
                    raise InvalidAPIParameters("Too large file")
                chunks.append(chunk)
                recv_size += chunk_size
            data = file.decode(b"".join(chunks))
            log.debug("received file: {0} ({1:,} bytes)", file.filename, recv_size)
            t = loop.create_task(self._agent_registry.upload_file(session, file.filename, data))
            upload_tasks.append(t)
        await asyncio.gather(*upload_tasks)

        return UploadFilesActionResult(result=None, session_row=session)

    async def modify_session(self, action: ModifySessionAction) -> ModifySessionActionResult:
        session_id = action.session_id
        props = action.modifier
        session_name = action.modifier.name.optional_value()

        async def _update(db_session: AsyncSession) -> Optional[SessionRow]:
            query_stmt = sa.select(SessionRow).where(SessionRow.id == session_id)
            session_row = await db_session.scalar(query_stmt)
            if session_row is None:
                raise ValueError(f"Session not found (id:{session_id})")
            session_row = cast(SessionRow, session_row)

            if session_name:
                # Check the owner of the target session has any session with the same name
                try:
                    sess = await SessionRow.get_session(
                        db_session,
                        session_name,
                        AccessKey(session_row.access_key),
                    )
                except SessionNotFound:
                    pass
                else:
                    raise ValueError(
                        f"Duplicate session name. Session(id:{sess.id}) already has the name"
                    )
            select_stmt = (
                sa.select(SessionRow)
                .options(selectinload(SessionRow.kernels))
                .execution_options(populate_existing=True)
                .where(SessionRow.id == session_id)
            )

            session_row = await db_session.scalar(select_stmt)
            to_update = props.fields_to_update()
            for key, value in to_update.items():
                setattr(session_row, key, value)

            if session_name:
                await db_session.execute(
                    sa.update(KernelRow)
                    .values(session_name=session_name)
                    .where(KernelRow.session_id == session_id)
                )
            return session_row

        async with self._db.connect() as db_conn:
            session_row = await execute_with_txn_retry(_update, self._db.begin_session, db_conn)
        if session_row is None:
            raise ValueError(f"Session not found (id:{session_id})")

        return ModifySessionActionResult(result=None, session_row=session_row)

    async def check_and_transit_status(
        self, action: CheckAndTransitStatusAction
    ) -> CheckAndTransitStatusActionResult:
        user_id = action.user_id
        user_role = action.user_role
        session_id = action.session_id

        async with self._db.begin_readonly_session() as db_session:
            session_row = await SessionRow.get_session_to_determine_status(db_session, session_id)
            if session_row.user_uuid != user_id and user_role not in (
                UserRole.ADMIN,
                UserRole.SUPERADMIN,
            ):
                log.warning(
                    f"You are not allowed to transit others's sessions status, skip (s:{session_id})"
                )
                return CheckAndTransitStatusActionResult(result={}, session_row=session_row)

        now = datetime.now(tzutc())
        session_rows = await self._agent_registry.session_lifecycle_mgr.transit_session_status(
            [session_id], now
        )
        await self._agent_registry.session_lifecycle_mgr.deregister_status_updatable_session([
            row.id for row, is_transited in session_rows if is_transited
        ])

        result = {row.id: row.status.name for row, _ in session_rows}
        return CheckAndTransitStatusActionResult(result=result, session_row=session_row)

    async def check_and_transit_status_multi(
        self, action: CheckAndTransitStatusBatchAction
    ) -> CheckAndTransitStatusBatchActionResult:
        user_id = action.user_id
        user_role = action.user_role
        session_ids = action.session_ids
        accessible_session_ids: list[SessionId] = []

        async with self._db.begin_readonly_session() as db_session:
            for sid in session_ids:
                session_row = await SessionRow.get_session_to_determine_status(db_session, sid)
                if session_row.user_uuid == user_id or user_role in (
                    UserRole.ADMIN,
                    UserRole.SUPERADMIN,
                ):
                    accessible_session_ids.append(sid)
                else:
                    log.warning(
                        f"You are not allowed to transit others's sessions status, skip (s:{sid})"
                    )

        now = datetime.now(tzutc())
        if accessible_session_ids:
            session_rows = await self._agent_registry.session_lifecycle_mgr.transit_session_status(
                accessible_session_ids, now
            )
            await self._agent_registry.session_lifecycle_mgr.deregister_status_updatable_session([
                row.id for row, is_transited in session_rows if is_transited
            ])
            result = {row.id: row.status.name for row, _ in session_rows}
        else:
            result = {}

        return CheckAndTransitStatusBatchActionResult(session_status_map=result)
